"""
Finite field elements implemented via PARI's FFELT type

AUTHORS:

- Peter Bruin (June 2013): initial version, based on
  element_ext_pari.py by William Stein et al. and
  element_ntl_gf2e.pyx by Martin Albrecht.
"""
# ****************************************************************************
#      Copyright (C) 2013 Peter Bruin <peter.bruin@math.uzh.ch>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from cypari2.paridecl cimport *
from cypari2.paripriv cimport *
from sage.libs.pari.convert_gmp cimport _new_GEN_from_mpz_t
from cypari2.stack cimport new_gen, new_gen_noclear, clear_stack
from cypari2.gen cimport Gen as pari_gen, objtogen

from sage.rings.finite_rings.element_base cimport FinitePolyExtElement
from sage.rings.finite_rings.integer_mod import IntegerMod_abstract

import sage.rings.integer
from sage.rings.integer cimport Integer
from sage.rings.polynomial.polynomial_element import Polynomial
from sage.rings.polynomial.multi_polynomial_element import MPolynomial
from sage.rings.rational import Rational
from sage.structure.element cimport Vector
from sage.structure.richcmp cimport rich_to_bool

from sage.interfaces.abc import GapElement


cdef GEN _INT_to_FFELT(GEN g, GEN x) except NULL:
    """
    Convert the t_INT `x` to an element of the field of definition of
    the t_FFELT `g`.

    This function must be called within ``sig_on()`` ... ``sig_off()``.

    TESTS:

    Converting large integers to finite field elements does not lead
    to overflow errors (see :issue:`16807`)::

        sage: p = previous_prime(2^64)
        sage: F.<x> = GF(p^2)
        sage: x * 2^63
        9223372036854775808*x
    """
    cdef GEN f, p = gel(g, 4), result
    cdef long t

    x = modii(x, p)
    if gequal0(x):
        return FF_zero(g)
    elif gequal1(x):
        return FF_1(g)
    else:
        # In characteristic 2, we have already dealt with the
        # two possible values of x, so we may assume that the
        # characteristic is > 2.
        t = g[1]  # codeword: t_FF_FpXQ, t_FF_Flxq, t_FF_F2xq
        if t == t_FF_FpXQ:
            f = cgetg(3, t_POL)
            set_gel(f, 1, gmael(g, 2, 1))
            set_gel(f, 2, x)
        elif t == t_FF_Flxq:
            f = cgetg(3, t_VECSMALL)
            set_gel(f, 1, gmael(g, 2, 1))
            f[2] = itou(x)
        else:
            sig_off()
            raise TypeError("unknown PARI finite field type")
        result = cgetg(5, t_FFELT)
        result[1] = t
        set_gel(result, 2, f)
        set_gel(result, 3, gel(g, 3))  # modulus
        set_gel(result, 4, p)
        return result

cdef class FiniteFieldElement_pari_ffelt(FinitePolyExtElement):
    """
    An element of a finite field implemented using PARI.

    EXAMPLES::

        sage: K = FiniteField(10007^10, 'a', implementation='pari_ffelt')
        sage: a = K.gen(); a
        a
        sage: type(a)
        <class 'sage.rings.finite_rings.element_pari_ffelt.FiniteFieldElement_pari_ffelt'>

    TESTS::

        sage: n = 63
        sage: m = 3
        sage: K.<a> = GF(2^n, implementation='pari_ffelt')
        sage: f = conway_polynomial(2, n)
        sage: f(a) == 0
        True
        sage: e = (2^n - 1) / (2^m - 1)
        sage: conway_polynomial(2, m)(a^e) == 0
        True

        sage: K.<a> = FiniteField(2^16, implementation='pari_ffelt')
        sage: K(0).is_zero()
        True
        sage: (a - a).is_zero()
        True
        sage: a - a
        0
        sage: a == a
        True
        sage: a - a == 0
        True
        sage: a - a == K(0)
        True
        sage: TestSuite(a).run()

    Test creating elements from basic Python types::

        sage: K.<a> = FiniteField(7^20, implementation='pari_ffelt')
        sage: K(int(8))
        1

    ::

        sage: k = FiniteField(3^4, 'a', implementation='pari_ffelt')
        sage: b = k(5) # indirect doctest
        sage: b.parent()
        Finite Field in a of size 3^4
        sage: a = k.gen()
        sage: k(a + 2)
        a + 2

    Univariate polynomials coerce into finite fields by evaluating
    the polynomial at the field's generator::

        sage: R.<x> = QQ[]
        sage: k.<a> = FiniteField(5^2, 'a', implementation='pari_ffelt')
        sage: k(R(2/3))
        4
        sage: k(x^2)
        a + 3

        sage: R.<x> = GF(5)[]
        sage: k(x^3-2*x+1)
        2*a + 4

        sage: x = polygen(QQ)
        sage: k(x^25)
        a

        sage: Q.<q> = FiniteField(5^7, 'q', implementation='pari_ffelt')
        sage: L = GF(5)
        sage: LL.<xx> = L[]
        sage: Q(xx^2 + 2*xx + 4)
        q^2 + 2*q + 4

        sage: k = FiniteField(3^11, 't', implementation='pari_ffelt')
        sage: k.polynomial()
        t^11 + 2*t^2 + 1
        sage: P = k.polynomial_ring()
        sage: k(P.0^11)
        t^2 + 2

    An element can be specified by its vector of coordinates with
    respect to the basis consisting of powers of the generator:

        sage: k = FiniteField(3^11, 't', implementation='pari_ffelt')
        sage: V = k.vector_space(map=False)
        sage: V
        Vector space of dimension 11 over Finite Field of size 3
        sage: v = V([0,1,2,0,1,2,0,1,2,0,1])
        sage: k(v)
        t^10 + 2*t^8 + t^7 + 2*t^5 + t^4 + 2*t^2 + t

    Multivariate polynomials only coerce if constant::

        sage: k = FiniteField(5^2, 'a', implementation='pari_ffelt')
        sage: R = k['x,y,z']; R
        Multivariate Polynomial Ring in x, y, z over Finite Field in a of size 5^2
        sage: k(R(2))
        2
        sage: R = QQ['x,y,z']
        sage: k(R(1/5))
        Traceback (most recent call last):
        ...
        ZeroDivisionError: inverse of Mod(0, 5) does not exist

    Gap elements can also be coerced into finite fields::

        sage: F = FiniteField(2^3, 'a', implementation='pari_ffelt')
        sage: a = F.multiplicative_generator(); a
        a
        sage: b = gap(a^3); b                                                           # needs sage.libs.gap
        Z(2^3)^3
        sage: F(b)
        a + 1
        sage: a^3
        a + 1

        sage: a = GF(13)(gap('0*Z(13)')); a                                             # needs sage.libs.gap
        0
        sage: a.parent()
        Finite Field of size 13

        sage: F = FiniteField(2^4, 'a', implementation='pari_ffelt')
        sage: F(gap('Z(16)^3'))                                                         # needs sage.libs.gap
        a^3
        sage: F(gap('Z(16)^2'))                                                         # needs sage.libs.gap
        a^2

    You can also call a finite extension field with a string
    to produce an element of that field, like this::

        sage: k = GF(2^8, 'a')
        sage: k('a^200')
        a^4 + a^3 + a^2

    This is especially useful for conversion from Singular etc.

    TESTS::

        sage: k = FiniteField(3^2, 'a', implementation='pari_ffelt')
        sage: a = k(11); a
        2
        sage: a.parent()
        Finite Field in a of size 3^2
        sage: V = k.vector_space(map=False); v = V((1,2))
        sage: k(v)
        2*a + 1

    We create elements using a list and verify that :issue:`10486` has
    been fixed::

        sage: k = FiniteField(3^11, 't', implementation='pari_ffelt')
        sage: x = k([1,0,2,1]); x
        t^3 + 2*t^2 + 1
        sage: x + x + x
        0
        sage: pari(x)
        t^3 + 2*t^2 + 1

    If the list is longer than the degree, we just get the result
    modulo the modulus::

        sage: from sage.rings.finite_rings.finite_field_pari_ffelt import FiniteField_pari_ffelt
        sage: R.<a> = PolynomialRing(GF(5))
        sage: k = FiniteField_pari_ffelt(5, a^2 - 2, 't')
        sage: x = k([0,0,0,1]); x
        2*t
        sage: pari(x)
        2*t

    When initializing from a list, the elements are first coerced
    to the prime field (:issue:`11685`)::

        sage: k = FiniteField(3^11, 't', implementation='pari_ffelt')
        sage: k([ 0, 1/2 ])
        2*t
        sage: k([ 0, 1/2, 0, 0, 0, 0, 0, 0, 0, -1, 0 ])
        2*t^9 + 2*t
        sage: k([ k(0), k(1) ])
        t
        sage: k([ GF(3)(2), GF(3^5,'u')(1) ])
        t + 2
        sage: R.<x> = PolynomialRing(k)
        sage: k([ x/x ])
        1
        sage: k([ R(-1), x/x ])
        t + 2
        sage: k([ R(-1), R(0), 0 ])
        2

    Check that zeros are created correctly (:issue:`11685`)::

        sage: K = FiniteField(3^11, 't', implementation='pari_ffelt'); a = K.0
        sage: v = 0; pari(K(v))
        0
        sage: v = Mod(0,3); pari(K(v))
        0
        sage: v = pari(0); pari(K(v))
        0
        sage: v = pari("Mod(0,3)"); pari(K(v))
        0
        sage: v = []; pari(K(v))
        0
        sage: v = [0]; pari(K(v))
        0
        sage: v = [0,0]; pari(K(v))
        0
        sage: v = pari("Pol(0)"); pari(K(v))
        0
        sage: v = pari("Mod(0, %s)"%K.modulus()); pari(K(v))
        0
        sage: v = pari("Mod(Pol(0), %s)"%K.modulus()); pari(K(v))
        0
        sage: v = K(1) - K(1); pari(K(v))
        0
        sage: v = K([1]) - K([1]); pari(K(v))
        0
        sage: v = a - a; pari(K(v))
        0
        sage: v = K(1)*0; pari(K(v))
        0
        sage: v = K([1])*K([0]); pari(K(v))
        0
        sage: v = a*0; pari(K(v))
        0
    """

    def __init__(self, parent, x):
        """
        Initialise ``self`` with the given ``parent`` and value
        converted from ``x``.

        This is called when constructing elements from Python.

        TESTS::

            sage: from sage.rings.finite_rings.element_pari_ffelt import FiniteFieldElement_pari_ffelt
            sage: K = FiniteField(101^2, 'a', implementation='pari_ffelt')
            sage: x = FiniteFieldElement_pari_ffelt(K, 'a + 1')
            sage: x
            a + 1
        """
        # FinitePolyExtElement.__init__(self, parent)
        self._parent = parent
        self.construct_from(x)

    def __dealloc__(self):
        """
        Cython destructor.
        """
        if self.val is not NULL:
            gunclone_deep(self.val)

    cdef FiniteFieldElement_pari_ffelt _new(self):
        """
        Create an empty element with the same parent as ``self``.
        """
        cdef FiniteFieldElement_pari_ffelt x
        x = FiniteFieldElement_pari_ffelt.__new__(FiniteFieldElement_pari_ffelt)
        x._parent = self._parent
        return x

    cdef void construct(self, GEN g) noexcept:
        """
        Initialise ``self`` to the FFELT ``g``, reset the PARI stack,
        and call sig_off().

        This should be called exactly once on every instance.
        """
        self.val = gcloneref(g)
        clear_stack()

    cdef int construct_from(self, x) except -1:
        """
        Initialise ``self`` to an FFELT constructed from the Sage
        object `x`.

        TESTS:

        Conversion of elements of the underlying vector space works in
        large characteristic (see :issue:`21186`)::

            sage: p = 13189065031705623239
            sage: Fq = FiniteField(p^3, "a")
            sage: Fq_X = PolynomialRing(Fq, "x")
            sage: pol = Fq_X("x^9 + 13189065031705622723*x^7 + 13189065031705622723*x^6 + 9288*x^5 + 18576*x^4 + 13189065031705590731*x^3 + 13189065031705497851*x^2 + 13189065031705497851*x + 13189065031705581443")
            sage: R = [r[0] for r in pol.roots()]
            sage: prod(Fq_X.gen() - r for r in R) == pol
            True
        """
        cdef GEN f, g, result, x_GEN
        cdef long i, n, t
        cdef Integer xi

        if isinstance(x, FiniteFieldElement_pari_ffelt):
            if self._parent is (<FiniteFieldElement_pari_ffelt>x)._parent:
                sig_on()
                self.construct((<FiniteFieldElement_pari_ffelt>x).val)
            else:
                raise TypeError("no coercion defined")

        elif isinstance(x, Integer):
            g = (<pari_gen>self._parent._gen_pari).g
            sig_on()
            x_GEN = _new_GEN_from_mpz_t((<Integer>x).value)
            self.construct(_INT_to_FFELT(g, x_GEN))

        elif isinstance(x, int):
            g = (<pari_gen>self._parent._gen_pari).g
            x = objtogen(x)
            sig_on()
            x_GEN = (<pari_gen>x).g
            self.construct(_INT_to_FFELT(g, x_GEN))

        elif isinstance(x, IntegerMod_abstract):
            if self._parent.characteristic().divides(x.modulus()):
                g = (<pari_gen>self._parent._gen_pari).g
                sig_on()
                x_GEN = _new_GEN_from_mpz_t(Integer(x).value)
                self.construct(_INT_to_FFELT(g, x_GEN))
            else:
                raise TypeError("no coercion defined")

        elif x is None:
            g = (<pari_gen>self._parent._gen_pari).g
            sig_on()
            self.construct(FF_zero(g))

        elif isinstance(x, pari_gen):
            g = (<pari_gen>self._parent._gen_pari).g
            x_GEN = (<pari_gen>x).g

            sig_on()
            if gequal0(x_GEN):
                self.construct(FF_zero(g))
                return 0
            elif gequal1(x_GEN):
                self.construct(FF_1(g))
                return 0

            t = typ(x_GEN)
            if t == t_FFELT:
                if FF_samefield(x_GEN, g):
                    self.construct(x_GEN)
                    return 0
            elif t == t_INT:
                self.construct(_INT_to_FFELT(g, x_GEN))
                return 0
            elif t == t_INTMOD:
                if gequal0(modii(gel(x_GEN, 1), FF_p_i(g))):
                    self.construct(_INT_to_FFELT(g, gel(x_GEN, 2)))
                    return 0
            elif t == t_FRAC:
                if not gequal0(modii(gel(x_GEN, 2), FF_p_i(g))):
                    elt = FF_div(_INT_to_FFELT(g, gel(x_GEN, 1)),
                                 _INT_to_FFELT(g, gel(x_GEN, 2)))
                    self.construct(elt)
                    return 0
            sig_off()
            raise TypeError(f"unable to convert PARI {x.type()} to finite field element")

        elif (isinstance(x, Vector)
              and x.parent() is self._parent.vector_space(map=False)):
            g = (<pari_gen>self._parent._gen_pari).g
            t = g[1]  # codeword: t_FF_FpXQ, t_FF_Flxq, t_FF_F2xq
            n = len(x)
            while n > 0 and x[n - 1] == 0:
                n -= 1
            sig_on()
            if n == 0:
                self.construct(FF_zero(g))
                return 0
            if t == t_FF_FpXQ:
                f = cgetg(n + 2, t_POL)
                set_gel(f, 1, gmael(g, 2, 1))
                for i in range(n):
                    xi = Integer(x[i])
                    set_gel(f, i + 2, _new_GEN_from_mpz_t(xi.value))
            elif t == t_FF_Flxq or t == t_FF_F2xq:
                f = cgetg(n + 2, t_VECSMALL)
                set_gel(f, 1, gmael(g, 2, 1))
                for i in range(n):
                    set_uel(f, i + 2, x[i])
                if t == t_FF_F2xq:
                    f = Flx_to_F2x(f)
            else:
                sig_off()
                raise TypeError("unknown PARI finite field type")
            result = cgetg(5, t_FFELT)
            result[1] = t
            set_gel(result, 2, f)
            set_gel(result, 3, gel(g, 3))  # modulus
            set_gel(result, 4, gel(g, 4))  # p
            self.construct(result)

        elif isinstance(x, Rational):
            self.construct_from(x % self._parent.characteristic())

        elif isinstance(x, Polynomial):
            if x.base_ring() is not self._parent.base_ring():
                x = x.change_ring(self._parent.base_ring())
            self.construct_from(x.substitute(self._parent.gen()))

        elif isinstance(x, MPolynomial) and x.is_constant():
            self.construct_from(x.constant_coefficient())

        elif isinstance(x, list):
            n = len(x)
            if n == 0:
                self.construct_from(None)
            elif n == 1:
                Fp = self._parent.base_ring()
                self.construct_from(Fp(x[0]))
            elif n == self._parent.degree():
                self.construct_from(self._parent.vector_space(map=False)(x))
            else:
                Fp = self._parent.base_ring()
                self.construct_from(self._parent.polynomial_ring()([Fp(y) for y in x]))

        elif isinstance(x, str):
            self.construct_from(self._parent.polynomial_ring()(x))

        elif isinstance(x, GapElement):
            try:
                from sage.libs.gap.libgap import libgap
                self.construct_from(libgap(x).sage(ring=self._parent))
            except (ValueError, IndexError, TypeError):
                raise TypeError("no coercion defined")

        elif isinstance(x, sage.libs.gap.element.GapElement_FiniteField):
            try:
                self.construct_from(x.sage(ring=self._parent))
            except (ValueError, IndexError, TypeError):
                raise TypeError("no coercion defined")

        else:
            raise TypeError("no coercion defined")

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: k.<c> = GF(3^17, implementation='pari_ffelt')
            sage: c^20  # indirect doctest
            c^4 + 2*c^3

        TESTS::

            sage: F.<I> = GF((2^128+51)^2)
            sage: I
            I
        """
        s = str(new_gen_noclear(self.val))
        if self._parent._need_replace_varname:
            s = s.replace("x", self._parent._names[0])  # .variable_name() is slower
        return s

    def __hash__(self):
        """
        Return the hash of ``self``.  This is by definition equal to
        the hash of ``self.polynomial()``.

        EXAMPLES::

            sage: k.<a> = GF(3^15, implementation='pari_ffelt')
            sage: R = GF(3)['a']; aa = R.gen()
            sage: hash(a^2 + 1) == hash(aa^2 + 1)
            True
        """
        return hash(self.polynomial())

    def __reduce__(self):
        """
        For pickling.

        TESTS::

            sage: K.<a> = FiniteField(10007^10, implementation='pari_ffelt')
            sage: loads(a.dumps()) == a
            True
        """
        return unpickle_FiniteFieldElement_pari_ffelt, (self._parent, str(self))

    def __copy__(self):
        """
        TESTS::

            sage: k.<a> = FiniteField(3^3, implementation='pari_ffelt')
            sage: a
            a
            sage: b = copy(a); b
            a
            sage: a is b
            True
        """
        # immutable
        return self

    def __deepcopy__(self, memo):
        """
        TESTS::

            sage: k.<a> = FiniteField(3^3, implementation='pari_ffelt')
            sage: a
            a
            sage: b = deepcopy(a); b
            a
            sage: a is b
            True
        """
        # immutable
        return self

    cpdef _richcmp_(self, other, int op):
        """
        Comparison of finite field elements.

        .. NOTE::

            Finite fields are unordered.  However, for the purpose of
            this function, we adopt the lexicographic ordering on the
            representing polynomials.

        EXAMPLES::

            sage: k.<a> = GF(2^20, implementation='pari_ffelt')
            sage: e = k.random_element()
            sage: f = loads(dumps(e))
            sage: e is f
            False
            sage: e == f
            True
            sage: e != (e + 1)
            True

        ::

            sage: K.<a> = GF(2^100, implementation='pari_ffelt')
            sage: a < a^2
            True
            sage: a > a^2
            False
            sage: a+1 > a^2
            False
            sage: a+1 < a^2
            True
            sage: a+1 < a
            False
            sage: a+1 == a
            False
            sage: a == a
            True

        TESTS::

            sage: k.<a> = FiniteField(3^3, implementation='pari_ffelt')
            sage: a == 1
            False
            sage: a^0 == 1
            True
            sage: a == a
            True
            sage: a < a^2
            True
            sage: a > a^2
            False
        """
        cdef int r
        sig_on()
        r = cmp_universal(self.val, (<FiniteFieldElement_pari_ffelt>other).val)
        sig_off()
        return rich_to_bool(op, r)

    cpdef _add_(self, right):
        """
        Addition.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: a + a^2 # indirect doctest
            a^2 + a
        """
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_add((<FiniteFieldElement_pari_ffelt>self).val,
                           (<FiniteFieldElement_pari_ffelt>right).val))
        return x

    cpdef _sub_(self, right):
        """
        Subtraction.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: a - a # indirect doctest
            0
        """
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_sub((<FiniteFieldElement_pari_ffelt>self).val,
                           (<FiniteFieldElement_pari_ffelt>right).val))
        return x

    cpdef _mul_(self, right):
        """
        Multiplication.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: (a^12 + 1)*(a^15 - 1) # indirect doctest
            a^15 + 2*a^12 + a^11 + 2*a^10 + 2
        """
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_mul((<FiniteFieldElement_pari_ffelt>self).val,
                           (<FiniteFieldElement_pari_ffelt>right).val))
        return x

    cpdef _div_(self, right):
        """
        Division.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: (a - 1) / (a + 1) # indirect doctest
            2*a^16 + a^15 + 2*a^14 + a^13 + 2*a^12 + a^11 + 2*a^10 + a^9 + 2*a^8 + a^7 + 2*a^6 + a^5 + 2*a^4 + a^3 + 2*a^2 + a + 1
        """
        if FF_equal0((<FiniteFieldElement_pari_ffelt>right).val):
            raise ZeroDivisionError
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_div((<FiniteFieldElement_pari_ffelt>self).val,
                           (<FiniteFieldElement_pari_ffelt>right).val))
        return x

    def is_zero(self):
        """
        Return ``True`` if ``self`` equals 0.

        EXAMPLES::

            sage: F.<a> = FiniteField(5^3, implementation='pari_ffelt')
            sage: a.is_zero()
            False
            sage: (a - a).is_zero()
            True
        """
        return bool(FF_equal0(self.val))

    def is_one(self):
        """
        Return ``True`` if ``self`` equals 1.

        EXAMPLES::

            sage: F.<a> = FiniteField(5^3, implementation='pari_ffelt')
            sage: a.is_one()
            False
            sage: (a/a).is_one()
            True
        """
        return bool(FF_equal1(self.val))

    def is_unit(self):
        """
        Return ``True`` if ``self`` is nonzero.

        EXAMPLES::

            sage: F.<a> = FiniteField(5^3, implementation='pari_ffelt')
            sage: a.is_unit()
            True
        """
        return not bool(FF_equal0(self.val))

    __bool__ = is_unit

    def __pos__(self):
        """
        Unitary positive operator...

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: +a
            a
        """
        return self

    def __neg__(self):
        """
        Negation.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: -a
            2*a
        """
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_neg_i((<FiniteFieldElement_pari_ffelt>self).val))
        return x

    def __invert__(self):
        """
        Return the multiplicative inverse of ``self``.

        EXAMPLES::

            sage: k.<a> = FiniteField(3^2, implementation='pari_ffelt')
            sage: ~a
            a + 2
            sage: (a+1)*a
            2*a + 1
            sage: ~((2*a)/a)
            2
        """
        if FF_equal0(self.val):
            raise ZeroDivisionError
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_inv((<FiniteFieldElement_pari_ffelt>self).val))
        return x

    def __pow__(FiniteFieldElement_pari_ffelt self, exp, other):
        """
        Exponentiation.

        TESTS::

            sage: K.<a> = GF(5^10, implementation='pari_ffelt')
            sage: n = (2*a)/a
            sage: n^-15
            2

        Large exponents are not a problem::

            sage: e = 3^10000
            sage: a^e
            2*a^9 + a^5 + 4*a^4 + 4*a^3 + a^2 + 3*a
            sage: a^(e % (5^10 - 1))
            2*a^9 + a^5 + 4*a^4 + 4*a^3 + a^2 + 3*a

        The exponent is converted to an integer (see :issue:`16540`)::

            sage: q = 11^23
            sage: F.<a> = FiniteField(q)
            sage: a^Mod(1, q - 1)
            a

        .. WARNING::

            For efficiency reasons, we do not verify that the
            exponentiation is well defined before converting the
            exponent to an integer.  This means that ``a^Mod(1, n)``
            returns `a` even if `n` is not a multiple of the
            multiplicative order of `a`.
        """
        if exp == 0:
            return self._parent.one()
        if exp < 0 and FF_equal0(self.val):
            raise ZeroDivisionError
        exp = Integer(exp).__pari__()
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(FF_pow(self.val, (<pari_gen>exp).g))
        return x

    def pth_power(FiniteFieldElement_pari_ffelt self, int k=1):
        r"""
        Return the `(p^k)`-th power of ``self``, where `p` is the
        characteristic of the field.

        INPUT:

        - ``k`` -- integer (default: 1); must fit in a C ``int``

        Note that if `k` is negative, then this computes the appropriate root.

        TESTS::

            sage: F.<a> = GF(13^64, implementation='pari_ffelt'); F
            Finite Field in a of size 13^64
            sage: x = F.random_element()
            sage: x.pth_power(0) == x
            True
            sage: x.pth_power(1) == x**13
            True
            sage: x.pth_power(2) == x**(13**2)
            True
            sage: x.pth_power(-1)**13 == x
            True

            sage: F.<a> = GF(127^16, implementation='pari_ffelt'); F
            Finite Field in a of size 127^16
            sage: x = F.random_element()
            sage: x.pth_power(0) == x
            True
            sage: x.pth_power(1) == x**127
            True
            sage: x.pth_power(2) == x**(127**2)
            True
            sage: x.pth_power(-1)**127 == x
            True
        """
        cdef int n = int(self._parent.degree())
        if k % n == 0:
            return self
        cdef Integer p = self._parent.characteristic()
        if k == 1 and (p < 100 or p.bit_length()**2 < n):
            # For extremely small primes or very large extension degrees,
            # exponentiation is faster.
            return self**p
        # Otherwise use PARI field morphism (evaluation of a Fp polynomial
        # at the image of the generator).
        f = self._parent._pari_frobenius(k)
        cdef FiniteFieldElement_pari_ffelt x = self._new()
        sig_on()
        x.construct(ffmap((<pari_gen>f).g, self.val))
        return x

    frobenius = pth_power

    def polynomial(self, name=None):
        """
        Return the unique representative of ``self`` as a polynomial
        over the prime field whose degree is less than the degree of
        the finite field over its prime field.

        INPUT:

        - ``name`` -- (optional) variable name

        EXAMPLES::

            sage: k.<a> = FiniteField(3^2, implementation='pari_ffelt')
            sage: pol = a.polynomial()
            sage: pol
            a
            sage: parent(pol)
            Univariate Polynomial Ring in a over Finite Field of size 3

        ::

            sage: k = FiniteField(3^4, 'alpha', implementation='pari_ffelt')
            sage: a = k.gen()
            sage: a.polynomial()
            alpha
            sage: (a**2 + 1).polynomial('beta')
            beta^2 + 1
            sage: (a**2 + 1).polynomial().parent()
            Univariate Polynomial Ring in alpha over Finite Field of size 3
            sage: (a**2 + 1).polynomial('beta').parent()
            Univariate Polynomial Ring in beta over Finite Field of size 3
        """
        sig_on()
        pol = new_gen(FF_to_FpXQ(self.val))
        return self._parent.polynomial_ring(name)(pol)

    def minpoly(self, var='x'):
        """
        Return the minimal polynomial of ``self``.

        INPUT:

        - ``var`` -- string (default: ``'x'``); variable name to use

        EXAMPLES::

            sage: R.<x> = PolynomialRing(FiniteField(3))
            sage: F.<a> = FiniteField(3^2, modulus=x^2 + 1, implementation='pari_ffelt')
            sage: a.minpoly('y')
            y^2 + 1
        """
        sig_on()
        pol = new_gen(FF_minpoly(self.val))
        return self._parent.polynomial_ring(var)(pol)

    def charpoly(self, var='x'):
        """
        Return the characteristic polynomial of ``self``.

        INPUT:

        - ``var`` -- string (default: ``'x'``); variable name to use

        EXAMPLES::

            sage: R.<x> = PolynomialRing(FiniteField(3))
            sage: F.<a> = FiniteField(3^2, modulus=x^2 + 1, implementation='pari_ffelt')
            sage: a.charpoly('y')
            y^2 + 1
        """
        sig_on()
        pol = new_gen(FF_charpoly(self.val))
        return self._parent.polynomial_ring(var)(pol)

    def is_square(self):
        """
        Return ``True`` if and only if ``self`` is a square in the
        finite field.

        EXAMPLES::

            sage: k.<a> = FiniteField(3^2, implementation='pari_ffelt')
            sage: a.is_square()
            False
            sage: (a**2).is_square()
            True

            sage: k.<a> = FiniteField(2^2, implementation='pari_ffelt')
            sage: (a**2).is_square()
            True

            sage: k.<a> = FiniteField(17^5, implementation='pari_ffelt')
            sage: (a**2).is_square()
            True
            sage: a.is_square()
            False
            sage: k(0).is_square()
            True
        """
        cdef long i
        sig_on()
        i = FF_issquare(self.val)
        sig_off()
        return bool(i)

    def sqrt(self, extend=False, all=False):
        """
        Return a square root of ``self``, if it exists.

        INPUT:

        - ``extend`` -- boolean (default: ``False``)

           .. WARNING::

               This option is not implemented.

        - ``all`` -- boolean (default: ``False``)

        OUTPUT:

        A square root of ``self``, if it exists.  If ``all`` is
        ``True``, a list containing all square roots of ``self``
        (of length zero, one or two) is returned instead.

        If ``extend`` is ``True``, a square root is chosen in an
        extension field if necessary.  If ``extend`` is ``False``, a
        :exc:`ValueError` is raised if the element is not a square in the
        base field.

        .. WARNING::

           The ``extend`` option is not implemented (yet).

        EXAMPLES::

            sage: F = FiniteField(7^2, 'a', implementation='pari_ffelt')
            sage: F(2).sqrt()
            4
            sage: F(3).sqrt() in (2*F.gen() + 6, 5*F.gen() + 1)
            True
            sage: F(3).sqrt()**2
            3
            sage: F(4).sqrt(all=True)
            [2, 5]

            sage: K = FiniteField(7^3, 'alpha', implementation='pari_ffelt')
            sage: K(3).sqrt()
            Traceback (most recent call last):
            ...
            ValueError: element is not a square
            sage: K(3).sqrt(all=True)
            []

            sage: K.<a> = GF(3^17, implementation='pari_ffelt')
            sage: (a^3 - a - 1).sqrt()
            a^16 + 2*a^15 + a^13 + 2*a^12 + a^10 + 2*a^9 + 2*a^8 + a^7 + a^6 + 2*a^5 + a^4 + 2*a^2 + 2*a + 2
        """
        if extend:
            raise NotImplementedError
        cdef GEN s
        cdef FiniteFieldElement_pari_ffelt x, mx
        sig_on()
        if FF_issquareall(self.val, &s):
            x = self._new()
            x.construct(s)
            if not all:
                return x
            elif gequal0(x.val) or self._parent.characteristic() == 2:
                return [x]
            else:
                sig_on()
                mx = self._new()
                mx.construct(FF_neg_i(x.val))
                return [x, mx]
        else:
            sig_off()
            if all:
                return []
            else:
                raise ValueError("element is not a square")

    def log(self, base, order=None, *, check=False):
        """
        Return a discrete logarithm of ``self`` with respect to the
        given base.

        INPUT:

        - ``base`` -- nonzero field element
        - ``order`` -- integer (optional), the order of the base
        - ``check`` -- boolean (default: ``False``); if set,
          test whether the given ``order`` is correct

        OUTPUT:

        An integer `x` such that ``self`` equals ``base`` raised to
        the power `x`.  If no such `x` exists, a :exc:`ValueError` is
        raised.

        EXAMPLES::

            sage: F.<g> = FiniteField(2^10, implementation='pari_ffelt')
            sage: b = g; a = g^37
            sage: a.log(b)
            37
            sage: b^37; a
            g^8 + g^7 + g^4 + g + 1
            g^8 + g^7 + g^4 + g + 1

        ::

            sage: F.<a> = FiniteField(5^2, implementation='pari_ffelt')
            sage: F(-1).log(F(2))
            2
            sage: F(1).log(a)
            0

        ::

            sage: p = 2^127-1
            sage: F.<t> = GF((p, 3))
            sage: elt = F.random_element()^(p^2+p+1)
            sage: (elt^2).log(elt, p-1)
            2

        Passing the ``order`` argument can lead to huge speedups when
        factoring the order of the entire unit group is expensive but
        the order of the base element is much smaller::

            sage: %timeit (elt^2).log(elt)       # not tested
            6.18 s ± 85 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
            sage: %timeit (elt^2).log(elt, p-1)  # not tested
            147 ms ± 1.39 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)

        Some cases where the logarithm is not defined or does not exist::

            sage: F.<a> = GF(3^10, implementation='pari_ffelt')
            sage: a.log(-1)
            Traceback (most recent call last):
            ...
            ArithmeticError: element a does not lie in group generated by 2
            sage: a.log(0)
            Traceback (most recent call last):
            ...
            ArithmeticError: discrete logarithm with base 0 is not defined
            sage: F(0).log(1)
            Traceback (most recent call last):
            ...
            ArithmeticError: discrete logarithm of 0 is not defined

        TESTS:

        An example for ``check=True``::

            sage: a = GF(101^5).primitive_element()
            sage: a.log(a, 10510100500, check=True)
            1
            sage: a.log(a, 5255050250, check=True)
            Traceback (most recent call last):
            ...
            ValueError: element does not have the provided order
        """
        base = self._parent(base)
        if self.is_zero():
            raise ArithmeticError("discrete logarithm of 0 is not defined")
        if base.is_zero():
            raise ArithmeticError("discrete logarithm with base 0 is not defined")

        # Compute the order of the base to check whether the element actually
        # lies in the group generated by the base. PARI may otherwise enter an
        # infinite loop.
        # We also have to specify the order of the base anyway as PARI assumes
        # by default that the base generates the entire multiplicative group.
        cdef GEN x, base_order
        sig_on()
        if order is None:
            base_order = FF_order((<FiniteFieldElement_pari_ffelt>base).val, NULL)
        else:
            if check:
                from sage.groups.generic import has_order
                if not has_order(base, order, '*'):
                    clear_stack()
                    raise ValueError('element does not have the provided order')
            base_order = _new_GEN_from_mpz_t((<Integer>order).value)
        if not gequal1(powgi(self.val, base_order)):
            clear_stack()
            raise ArithmeticError(f'element {self} does not lie in group generated by {base}')
        x = FF_log(self.val, (<FiniteFieldElement_pari_ffelt>base).val, base_order)
        return Integer(new_gen(x))

    def multiplicative_order(self):
        """
        Return the order of ``self`` in the multiplicative group.

        EXAMPLES::

            sage: a = FiniteField(5^3, 'a', implementation='pari_ffelt').0
            sage: a.multiplicative_order()
            124
            sage: a**124
            1
        """
        if self.is_zero():
            raise ArithmeticError("Multiplicative order of 0 not defined.")
        cdef GEN order
        sig_on()
        order = FF_order(self.val, NULL)
        return Integer(new_gen(order))

    def lift(self):
        """
        If ``self`` is an element of the prime field, return a lift of
        this element to an integer.

        EXAMPLES::

            sage: k = FiniteField(next_prime(10^10)^2, 'u', implementation='pari_ffelt')
            sage: a = k(17)/k(19)
            sage: b = a.lift(); b
            7894736858
            sage: b.parent()
            Integer Ring
        """
        if FF_equal0(self.val):
            return Integer(0)
        f = self.polynomial()
        if f.degree() == 0:
            return f.constant_coefficient().lift()
        else:
            raise ValueError("element is not in the prime field")

    def _integer_(self, ZZ=None):
        """
        Lift to a Sage integer, if possible.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: b = k(2)
            sage: b._integer_()
            2
            sage: a._integer_()
            Traceback (most recent call last):
            ...
            ValueError: element is not in the prime field
        """
        return self.lift()

    def __int__(self):
        """
        Lift to a python int, if possible.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: b = k(2)
            sage: int(b)
            2
            sage: int(a)
            Traceback (most recent call last):
            ...
            ValueError: element is not in the prime field
        """
        return int(self.lift())

    def __float__(self):
        """
        Lift to a python float, if possible.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: b = k(2)
            sage: float(b)
            2.0
        """
        return float(self.lift())

    def __pari__(self, var=None):
        """
        Return a PARI object representing ``self``.

        EXAMPLES::

            sage: k.<a> = FiniteField(3^3, implementation='pari_ffelt')
            sage: b = a**2 + 2*a + 1
            sage: b.__pari__()
            a^2 + 2*a + 1

        TESTS::

            sage: F.<I> = GF((2^128+51)^2)
            sage: pari(I)
            Traceback (most recent call last):
            ...
            ValueError: variable name illegal in PARI
        """
        if self._parent._need_replace_varname:
            raise ValueError("variable name illegal in PARI")
        return new_gen_noclear(self.val)

    def _pari_init_(self):
        """
        Return a string representing ``self`` in PARI.

        EXAMPLES::

            sage: k.<a> = GF(3^17, implementation='pari_ffelt')
            sage: a._pari_init_()
            'subst(a+3*a,a,ffgen(Mod(1, 3)*x^17 + Mod(2, 3)*x + Mod(1, 3),a))'
            sage: k(1)._pari_init_()
            'subst(1+3*a,a,ffgen(Mod(1, 3)*x^17 + Mod(2, 3)*x + Mod(1, 3),a))'

        This is used for conversion to GP. The element is displayed
        as "a" but has correct arithmetic::

            sage: gp(a)
            a
            sage: gp(a).type()
            t_FFELT
            sage: gp(a)^100
            2*a^16 + 2*a^15 + a^4 + a + 1
            sage: gp(a^100)
            2*a^16 + 2*a^15 + a^4 + a + 1
            sage: gp(k(0))
            0
            sage: gp(k(0)).type()
            t_FFELT
        """
        ffgen = "ffgen(%s,a)" % self._parent.modulus()._pari_init_()
        # Add this "zero" to ensure that the polynomial is not constant
        zero = "%s*a" % self._parent.characteristic()
        return "subst(%s+%s,a,%s)" % (self, zero, ffgen)

    def _magma_init_(self, magma) -> str:
        """
        Return a string representing ``self`` in Magma.

        EXAMPLES::

            sage: GF(7)(3)._magma_init_(magma)            # optional - magma
            'GF(7)!3'
        """
        k = self._parent
        km = magma(k)
        return str(self).replace(k.variable_name(), km.gen(1).name())

    def _gap_init_(self) -> str:
        r"""
        Return the string representing ``self`` in GAP.

        .. NOTE::

            The order of the parent field must be `\leq 65536`.  This
            function can be slow since elements of non-prime finite
            fields are represented in GAP as powers of a generator for
            the multiplicative group, so a discrete logarithm must be
            computed.

        EXAMPLES::

            sage: F = FiniteField(2^3, 'aa', implementation='pari_ffelt')
            sage: aa = F.multiplicative_generator()
            sage: gap(aa)  # indirect doctest
            Z(2^3)
            sage: b = F.multiplicative_generator()
            sage: a = b^3
            sage: gap(a)
            Z(2^3)^3
            sage: gap(a^3)
            Z(2^3)^2
            sage: F(gap('Z(8)^3'))
            aa + 1
            sage: F(libgap.Z(8)^3)
            aa + 1

        You can specify the instance of the Gap interpreter that is used::

            sage: F = FiniteField(next_prime(200)^2, 'a', implementation='pari_ffelt')
            sage: a = F.multiplicative_generator()
            sage: a._gap_(gap)
            Z(211^2)
            sage: (a^20)._gap_(gap)
            Z(211^2)^20

        Gap only supports relatively small finite fields::

            sage: F = FiniteField(next_prime(1000)^2, 'a', implementation='pari_ffelt')
            sage: a = F.multiplicative_generator()
            sage: a._gap_init_()
            Traceback (most recent call last):
            ...
            TypeError: order must be at most 65536
            sage: gap.coerce(a)
            Traceback (most recent call last):
            ...
            TypeError: no canonical coercion from Finite Field in a of size 1009^2 to Gap
        """
        F = self._parent
        if F.order() > 65536:
            raise TypeError("order must be at most 65536")

        if self == 0:
            return '0*Z(%s)' % F.order()
        assert F.degree() > 1
        g = F.multiplicative_generator()
        n = self.log(g)
        return 'Z(%s)^%s' % (F.order(), n)


def unpickle_FiniteFieldElement_pari_ffelt(parent, elem):
    """
    EXAMPLES::

        sage: k.<a> = GF(2^20, implementation='pari_ffelt')
        sage: e = k.random_element()
        sage: f = loads(dumps(e)) # indirect doctest
        sage: e == f
        True
    """
    return parent(elem)



r"""
Eta-products on modular curves `X_0(N)`

This package provides a class for representing eta-products, which
are meromorphic functions on modular curves of the form

.. MATH::

    \prod_{d | N} \eta(q^d)^{r_d}

where `\eta(q)` is Dirichlet's eta function

.. MATH::

    q^{1/24} \prod_{n = 1}^\infty(1-q^n) .

These are useful for obtaining explicit models of modular curves.

See :issue:`3934` for background.

AUTHOR:

- David Loeffler (2008-08-22): initial version
"""

# ***************************************************************************
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#                     2008 David Loeffler <d.loeffler.01@cantab.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ***************************************************************************
from __future__ import annotations
from typing import Any

from sage.arith.misc import divisors, prime_divisors, euler_phi, is_square, gcd
from sage.categories.groups import Groups
from sage.matrix.constructor import matrix
from sage.modules.free_module import FreeModule
from sage.rings.finite_rings.integer_mod import Mod
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.power_series_ring import PowerSeriesRing
from sage.rings.rational_field import QQ
from sage.structure.element import Element
from sage.structure.formal_sum import FormalSum
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp, richcmp_method, op_EQ, op_NE
from sage.structure.sage_object import SageObject
from sage.structure.unique_representation import UniqueRepresentation

import weakref

_cache: dict[int, Any] = {}


def EtaGroup(level):
    r"""
    Create the group of eta products of the given level.

    EXAMPLES::

        sage: EtaGroup(12)
        Group of eta products on X_0(12)
        sage: EtaGroup(1/2)
        Traceback (most recent call last):
        ...
        TypeError: Level (=1/2) must be a positive integer
        sage: EtaGroup(0)
        Traceback (most recent call last):
        ...
        ValueError: Level (=0) must be a positive integer
    """
    if level in _cache:
        G = _cache[level]()
        if G is not None:
            return G
    G = EtaGroup_class(level)
    _cache[level] = weakref.ref(G)
    return G


class EtaGroupElement(Element):

    def __init__(self, parent, rdict) -> None:
        r"""
        Create an eta product object. Usually called implicitly via
        EtaGroup_class.__call__ or the EtaProduct factory function.

        EXAMPLES::

            sage: EtaProduct(8, {1:24, 2:-24})
            Eta product of level 8 : (eta_1)^24 (eta_2)^-24
            sage: g = _; g == loads(dumps(g))
            True
            sage: TestSuite(g).run()
        """
        self._N = parent._N
        N = self._N

        if isinstance(rdict, EtaGroupElement):
            rdict = rdict._rdict
            # Note: This is needed because the "x in G" test tries to call G(x)
            # and see if it returns an error. So sometimes this will be getting
            # called with rdict being an eta product, not a dictionary.

        if rdict == 1:
            rdict = {}

        # Check Ligozat criteria
        sumR = sumDR = sumNoverDr = 0
        prod = 1
        for d in list(rdict):
            if N % d:
                raise ValueError("%s does not divide %s" % (d, N))

            if rdict[d] == 0:
                del rdict[d]
                continue
            sumR += rdict[d]
            sumDR += rdict[d] * d
            sumNoverDr += rdict[d] * (N // d)
            prod *= (N // d)**rdict[d]

        if sumR != 0:
            raise ValueError("sum r_d (=%s) is not 0" % sumR)
        if sumDR % 24:
            raise ValueError("sum d r_d (=%s) is not 0 mod 24" % sumDR)
        if sumNoverDr % 24:
            raise ValueError("sum (N/d) r_d (=%s) is not 0 mod 24" % sumNoverDr)
        if not is_square(prod):
            raise ValueError("product (N/d)^(r_d) (=%s) is not a square" % prod)

        self._sumDR = ZZ(sumDR)  # this is useful to have around
        self._rdict = rdict

        Element.__init__(self, parent)

    def _mul_(self, other):
        r"""
        Return the product of ``self`` and ``other``.

        EXAMPLES::

            sage: eta1, eta2 = EtaGroup(4).basis() # indirect doctest
            sage: eta1 * eta2
            Eta product of level 4 : (eta_1)^24 (eta_2)^-48 (eta_4)^24
        """
        newdict = {d: self._rdict.get(d, 0) + other._rdict.get(d, 0)
                   for d in set(self._rdict).union(other._rdict)}
        P = self.parent()
        return P.element_class(P, newdict)

    def _div_(self, other):
        r"""
        Return `self * other^{-1}`.

        EXAMPLES::

            sage: eta1, eta2 = EtaGroup(4).basis()
            sage: eta1 / eta2 # indirect doctest
            Eta product of level 4 : (eta_1)^-8 (eta_4)^8
            sage: (eta1 / eta2) * eta2 == eta1
            True
        """
        newdict = {d: self._rdict.get(d, 0) - other._rdict.get(d, 0)
                   for d in set(self._rdict).union(other._rdict)}
        P = self.parent()
        return P.element_class(P, newdict)

    def __invert__(self):
        r"""
        Return the inverse of ``self``.

        EXAMPLES::

            sage: eta1, eta2 = EtaGroup(4).basis()
            sage: ~eta2  # indirect doctest
            Eta product of level 4 : (eta_1)^-16 (eta_2)^24 (eta_4)^-8
        """
        newdict = {d: -self._rdict[d] for d in self._rdict}
        P = self.parent()
        return P.element_class(P, newdict)

    def is_one(self) -> bool:
        r"""
        Return whether ``self`` is the one of the monoid.

        EXAMPLES::

            sage: e = EtaProduct(3, {3:12, 1:-12})
            sage: e.is_one()
            False
            sage: e.parent().one().is_one()
            True
            sage: ep = EtaProduct(5, {})
            sage: ep.is_one()
            True
            sage: ep.parent().one() == ep
            True
        """
        return not self._rdict

    def _richcmp_(self, other, op) -> bool:
        r"""
        Compare ``self`` to ``other``.

        Eta products are compared according to their rdicts.

        EXAMPLES::

            sage: EtaProduct(2, {2:24,1:-24}) == 1
            False
            sage: EtaProduct(6, {1:-24, 2:24}) == EtaProduct(6, {1:-24, 2:24})
            True
            sage: EtaProduct(6, {1:-24, 2:24}) == EtaProduct(6, {1:24, 2:-24})
            False
            sage: EtaProduct(6, {1:-24, 2:24}) < EtaProduct(6, {1:-24, 2:24, 3:24, 6:-24})
            True
            sage: EtaProduct(6, {1:-24, 2:24, 3:24, 6:-24}) < EtaProduct(6, {1:-24, 2:24})
            False
        """
        if op in [op_EQ, op_NE]:
            test = (self._N == other._N and
                    self._rdict == other._rdict)
            return test == (op == op_EQ)
        return richcmp((self._N, sorted(self._rdict.items())),
                       (other._N, sorted(other._rdict.items())), op)

    def _short_repr(self) -> str:
        r"""
        A short string representation of ``self``, which does not specify the
        level.

        EXAMPLES::

            sage: EtaProduct(3, {3:12, 1:-12})._short_repr()
            '(eta_1)^-12 (eta_3)^12'
        """
        if self.degree() == 0:
            return "1"
        return " ".join("(eta_%s)^%s" % (d, exp)
                        for d, exp in sorted(self._rdict.items()))

    def _repr_(self) -> str:
        r"""
        Return the string representation of ``self``.

        EXAMPLES::

            sage: EtaProduct(3, {3:12, 1:-12})._repr_()
            'Eta product of level 3 : (eta_1)^-12 (eta_3)^12'
        """
        return "Eta product of level %s : " % self.level() + self._short_repr()

    def level(self) -> Integer:
        r"""
        Return the level of this eta product.

        EXAMPLES::

            sage: e = EtaProduct(3, {3:12, 1:-12})
            sage: e.level()
            3
            sage: EtaProduct(12, {6:6, 2:-6}).level() # not the lcm of the d's
            12
            sage: EtaProduct(36, {6:6, 2:-6}).level() # not minimal
            36
        """
        return self._N

    def q_expansion(self, n):
        r"""
        Return the `q`-expansion of ``self`` at the cusp at infinity.

        INPUT:

        - ``n`` -- integer; number of terms to calculate

        OUTPUT:

        A power series over `\ZZ` in the variable `q`, with a *relative*
        precision of `1 + O(q^n)`.

        ALGORITHM: Calculates eta to (n/m) terms, where m is the smallest
        integer dividing self.level() such that self.r(m) != 0. Then
        multiplies.

        EXAMPLES::

            sage: EtaProduct(36, {6:6, 2:-6}).q_expansion(10)
            q + 6*q^3 + 27*q^5 + 92*q^7 + 279*q^9 + O(q^11)
            sage: R.<q> = ZZ[[]]
            sage: EtaProduct(2,{2:24,1:-24}).q_expansion(100) == delta_qexp(101)(q^2)/delta_qexp(101)(q)
            True
        """
        R, q = PowerSeriesRing(ZZ, 'q').objgen()
        pr = R.one().O(n)
        if not self._rdict:  # if self.is_one():
            return pr
        # self.r(d) should always be nonzero since we filtered out the 0s
        eta_n = max(n // d for d in self._rdict)  # if self.r(d)
        eta = qexp_eta(R, eta_n)
        for d in self._rdict:
            rd = self._rdict[d]
            if rd:
                pr *= eta(q ** d) ** ZZ(rd)
        return pr * q**(self._sumDR // 24)

    def qexp(self, n):
        """
        Alias for ``self.q_expansion()``.

        EXAMPLES::

            sage: e = EtaProduct(36, {6:8, 3:-8})
            sage: e.qexp(10)
            q + 8*q^4 + 36*q^7 + O(q^10)
            sage: e.qexp(30) == e.q_expansion(30)
            True
        """
        return self.q_expansion(n)

    def order_at_cusp(self, cusp: CuspFamily) -> Integer:
        r"""
        Return the order of vanishing of ``self`` at the given cusp.

        INPUT:

        - ``cusp`` -- a :class:`CuspFamily` object

        OUTPUT: integer

        EXAMPLES::

            sage: e = EtaProduct(2, {2:24, 1:-24})
            sage: e.order_at_cusp(CuspFamily(2, 1)) # cusp at infinity
            1
            sage: e.order_at_cusp(CuspFamily(2, 2)) # cusp 0
            -1
        """
        if not isinstance(cusp, CuspFamily):
            raise TypeError("argument (=%s) should be a CuspFamily" % cusp)
        if cusp.level() != self._N:
            raise ValueError("cusp not on right curve")
        sigma = sum(ell * self._rdict[ell] / cusp.width() *
                    (gcd(cusp.width(), self._N // ell))**2
                    for ell in self._rdict)
        return sigma / ZZ(24) / gcd(cusp.width(), self._N // cusp.width())

    def divisor(self):
        r"""
        Return the divisor of ``self``, as a formal sum of CuspFamily objects.

        EXAMPLES::

            sage: e = EtaProduct(12, {1:-336, 2:576, 3:696, 4:-216, 6:-576, 12:-144})
            sage: e.divisor() # FormalSum seems to print things in a random order?
            -131*(Inf) - 50*(c_{2}) + 11*(0) + 50*(c_{6}) + 169*(c_{4}) - 49*(c_{3})
            sage: e = EtaProduct(2^8, {8:1,32:-1})
            sage: e.divisor() # random
            -(c_{2}) - (Inf) - (c_{8,2}) - (c_{8,3}) - (c_{8,4}) - (c_{4,2})
             - (c_{8,1}) - (c_{4,1}) + (c_{32,4}) + (c_{32,3}) + (c_{64,1})
             + (0) + (c_{32,2}) + (c_{64,2}) + (c_{128}) + (c_{32,1})
        """
        return FormalSum([(self.order_at_cusp(c), c)
                          for c in AllCusps(self.level())])

    def degree(self) -> Integer:
        r"""
        Return the degree of ``self`` as a map `X_0(N) \to \mathbb{P}^1`.

        This is the sum of all the positive coefficients in the divisor
        of ``self``.

        EXAMPLES::

            sage: e = EtaProduct(12, {1:-336, 2:576, 3:696, 4:-216, 6:-576, 12:-144})
            sage: e.degree()
            230
        """
        return sum(self.order_at_cusp(c)
                   for c in AllCusps(self._N)
                   if self.order_at_cusp(c) > 0)

    def r(self, d) -> Integer:
        r"""
        Return the exponent `r_d` of `\eta(q^d)` in ``self``.

        EXAMPLES::

            sage: e = EtaProduct(12, {2:24, 3:-24})
            sage: e.r(3)
            -24
            sage: e.r(4)
            0
        """
        return self._rdict.get(d, 0)


class EtaGroup_class(UniqueRepresentation, Parent):
    r"""
    The group of eta products of a given level under multiplication.

    TESTS::

        sage: TestSuite(EtaGroup(12)).run()

        sage: EtaGroup(12) == EtaGroup(12)
        True
        sage: EtaGroup(12) == EtaGroup(13)
        False

        sage: EtaGroup(12) != EtaGroup(12)
        False
        sage: EtaGroup(12) != EtaGroup(13)
        True
    """

    def __init__(self, level) -> None:
        r"""
        Create the group of eta products of a given level, which must be a
        positive integer.

        EXAMPLES::

            sage: G = EtaGroup(12); G # indirect doctest
            Group of eta products on X_0(12)
            sage: TestSuite(G).run()
        """
        try:
            level = ZZ(level)
        except TypeError:
            raise TypeError("Level (=%s) must be a positive integer" % level)
        if level < 1:
            raise ValueError("Level (=%s) must be a positive integer" % level)
        self._N = level
        Parent.__init__(self, category=Groups().Commutative())

    def _repr_(self) -> str:
        r"""
        String representation of ``self``.

        EXAMPLES::

            sage: EtaGroup(12)._repr_()
            'Group of eta products on X_0(12)'
        """
        return "Group of eta products on X_0(%s)" % self.level()

    def one(self) -> EtaGroupElement:
        r"""
        Return the identity element of ``self``.

        EXAMPLES::

            sage: EtaGroup(12).one()
            Eta product of level 12 : 1
        """
        return self({})

    def _element_constructor_(self, dic):
        r"""
        Create an element of this group (an eta product object) with
        exponents from the given dictionary.

        INPUT:

        - ``dic`` -- dictionary

        See the docstring of :func:`EtaProduct` for how ``dic`` is used.

        EXAMPLES::

            sage: EtaGroup(2).__call__({1:24, 2:-24})
            Eta product of level 2 : (eta_1)^24 (eta_2)^-24
        """
        return self.element_class(self, dic)

    def level(self) -> Integer:
        r"""
        Return the level of ``self``.

        EXAMPLES::

            sage: EtaGroup(10).level()
            10
        """
        return self._N

    def basis(self, reduce=True) -> list:
        r"""
        Produce a basis for the free abelian group of eta-products of level
        N (under multiplication), attempting to find basis vectors of the
        smallest possible degree.

        INPUT:

        - ``reduce`` -- boolean (default: ``True``); whether or not to apply
          LLL-reduction to the calculated basis

        EXAMPLES::

            sage: EtaGroup(5).basis()
            [Eta product of level 5 : (eta_1)^6 (eta_5)^-6]
            sage: EtaGroup(12).basis()
            [Eta product of level 12 : (eta_1)^-3 (eta_2)^2 (eta_3)^1 (eta_4)^-1 (eta_6)^-2 (eta_12)^3,
             Eta product of level 12 : (eta_1)^-4 (eta_2)^2 (eta_3)^4 (eta_6)^-2,
             Eta product of level 12 : (eta_1)^6 (eta_2)^-9 (eta_3)^-2 (eta_4)^3 (eta_6)^3 (eta_12)^-1,
             Eta product of level 12 : (eta_1)^-1 (eta_2)^3 (eta_3)^3 (eta_4)^-2 (eta_6)^-9 (eta_12)^6,
             Eta product of level 12 : (eta_1)^3 (eta_3)^-1 (eta_4)^-3 (eta_12)^1]
            sage: EtaGroup(12).basis(reduce=False) # much bigger coefficients
            [Eta product of level 12 : (eta_1)^384 (eta_2)^-576 (eta_3)^-696 (eta_4)^216 (eta_6)^576 (eta_12)^96,
             Eta product of level 12 : (eta_2)^24 (eta_12)^-24,
             Eta product of level 12 : (eta_1)^-40 (eta_2)^116 (eta_3)^96 (eta_4)^-30 (eta_6)^-80 (eta_12)^-62,
             Eta product of level 12 : (eta_1)^-4 (eta_2)^-33 (eta_3)^-4 (eta_4)^1 (eta_6)^3 (eta_12)^37,
             Eta product of level 12 : (eta_1)^15 (eta_2)^-24 (eta_3)^-29 (eta_4)^9 (eta_6)^24 (eta_12)^5]

        ALGORITHM: An eta product of level `N` is uniquely
        determined by the integers `r_d` for `d | N` with
        `d < N`, since `\sum_{d | N} r_d = 0`. The valid
        `r_d` are those that satisfy two congruences modulo 24,
        and one congruence modulo 2 for every prime divisor of N. We beef
        up the congruences modulo 2 to congruences modulo 24 by multiplying
        by 12. To calculate the kernel of the ensuing map
        `\ZZ^m \to (\ZZ/24\ZZ)^n`
        we lift it arbitrarily to an integer matrix and calculate its Smith
        normal form. This gives a basis for the lattice.

        This lattice typically contains "large" elements, so by default we
        pass it to the reduce_basis() function which performs
        LLL-reduction to give a more manageable basis.
        """
        N = self.level()
        divs = divisors(N)[:-1]
        s = len(divs)
        primedivs = prime_divisors(N)

        rows = []
        for di in divs:
            # generate a row of relation matrix
            row = [Mod(di, 24) - Mod(N, 24), Mod(N // di, 24) - Mod(1, 24)]
            row.extend(Mod(12 * (N // di).valuation(p), 24)
                       for p in primedivs)
            rows.append(row)

        M = matrix(IntegerModRing(24), rows)
        Mlift = M.change_ring(ZZ)
        # now we compute elementary factors of Mlift
        S, U, V = Mlift.smith_form()
        good_vects = []
        for i in range(U.nrows()):
            vect = U.row(i)
            nf = (i < S.ncols() and S[i, i]) or 0  # ?
            good_vects.append((vect * 24 / gcd(nf, 24)).list())
        for v in good_vects:
            v.append(-sum(list(v)))
        dicts: list[dict] = []
        for v in good_vects:
            dicts.append({})
            for i in range(s):
                dicts[-1][divs[i]] = v[i]
            dicts[-1][N] = v[-1]
        if reduce:
            return self.reduce_basis([self(d) for d in dicts])
        return [self(d) for d in dicts]

    def reduce_basis(self, long_etas) -> list:
        r"""
        Produce a more manageable basis via LLL-reduction.

        INPUT:

        - ``long_etas`` -- a list of EtaGroupElement objects (which
          should all be of the same level)

        OUTPUT:

        - a new list of EtaGroupElement objects having
          hopefully smaller norm

        ALGORITHM: We define the norm of an eta-product to be the
        `L^2` norm of its divisor (as an element of the free
        `\ZZ`-module with the cusps as basis and the
        standard inner product). Applying LLL-reduction to this gives a
        basis of hopefully more tractable elements. Of course we'd like to
        use the `L^1` norm as this is just twice the degree, which
        is a much more natural invariant, but `L^2` norm is easier
        to work with!

        EXAMPLES::

            sage: EtaGroup(4).reduce_basis([ EtaProduct(4, {1:8,2:24,4:-32}), EtaProduct(4, {1:8, 4:-8})])
            [Eta product of level 4 : (eta_1)^8 (eta_4)^-8,
             Eta product of level 4 : (eta_1)^-8 (eta_2)^24 (eta_4)^-16]
        """
        N = self.level()
        cusps = AllCusps(N)
        r = matrix(ZZ, [[et.order_at_cusp(c) for c in cusps] for et in long_etas])
        V = FreeModule(ZZ, r.ncols())
        A = V.submodule_with_basis([V(rw) for rw in r.rows()])
        rred = r.LLL()
        short_etas = []
        for shortvect in rred.rows():
            bv = A.coordinates(shortvect)
            dic = {d: sum(bv[i] * long_etas[i].r(d)
                          for i in range(r.nrows()))
                   for d in divisors(N)}
            short_etas.append(self(dic))
        return short_etas

    Element = EtaGroupElement


def EtaProduct(level, dic) -> EtaGroupElement:
    r"""
    Create an :class:`EtaGroupElement` object representing the function
    `\prod_{d | N} \eta(q^d)^{r_d}`.

    This checks the criteria of Ligozat to ensure that this product
    really is the `q`-expansion of a meromorphic function on `X_0(N)`.

    INPUT:

    - ``level`` -- integer; the N such that this eta
      product is a function on `X_0(N)`

    - ``dic`` -- a dictionary indexed by divisors of N such that the
      coefficient of `\eta(q^d)` is r[d]. Only nonzero coefficients need be
      specified. If Ligozat's criteria are not satisfied, a :exc:`ValueError`
      will be raised.

    OUTPUT:

    An EtaGroupElement object, whose parent is the EtaGroup of level N and
    whose coefficients are the given dictionary.

    .. NOTE::

        The dictionary ``dic`` does not uniquely specify N. It is
        possible for two EtaGroupElements with different `N`'s to
        be created with the same dictionary, and these represent different
        objects (although they will have the same `q`-expansion at
        the cusp `\infty`).

    EXAMPLES::

        sage: EtaProduct(3, {3:12, 1:-12})
        Eta product of level 3 : (eta_1)^-12 (eta_3)^12
        sage: EtaProduct(3, {3:6, 1:-6})
        Traceback (most recent call last):
        ...
        ValueError: sum d r_d (=12) is not 0 mod 24
        sage: EtaProduct(3, {4:6, 1:-6})
        Traceback (most recent call last):
        ...
        ValueError: 4 does not divide 3
    """
    return EtaGroup(level)(dic)


def n_cusps_of_width(N, d) -> Integer:
    r"""
    Return the number of cusps on `X_0(N)` of width ``d``.

    INPUT:

    - ``N`` -- integer; the level

    - ``d`` -- integer; an integer dividing `N`, the cusp width

    EXAMPLES::

        sage: from sage.modular.etaproducts import n_cusps_of_width
        sage: [n_cusps_of_width(18,d) for d in divisors(18)]
        [1, 1, 2, 2, 1, 1]
        sage: n_cusps_of_width(4,8)
        Traceback (most recent call last):
        ...
        ValueError: N and d must be positive integers with d|N

    TESTS:

    The old method name is kept as an alias::

        sage: from sage.modular.etaproducts import num_cusps_of_width
        sage: [num_cusps_of_width(6,d) for d in divisors(6)]
        [1, 1, 1, 1]
    """
    N = ZZ(N)
    d = ZZ(d)
    if N <= 0 or d <= 0 or N % d:
        raise ValueError("N and d must be positive integers with d|N")

    return euler_phi(d.gcd(N // d))


num_cusps_of_width = n_cusps_of_width


def AllCusps(N) -> list:
    r"""
    Return a list of :class:`CuspFamily` objects corresponding to the cusps of
    `X_0(N)`.

    INPUT:

    - ``N`` -- integer; the level

    EXAMPLES::

        sage: AllCusps(18)
        [(Inf), (c_{2}), (c_{3,1}), (c_{3,2}), (c_{6,1}), (c_{6,2}), (c_{9}), (0)]
        sage: AllCusps(0)
        Traceback (most recent call last):
        ...
        ValueError: N must be positive
    """
    N = ZZ(N)
    if N <= 0:
        raise ValueError("N must be positive")

    c = []
    for d in divisors(N):
        n = n_cusps_of_width(N, d)
        if n == 1:
            c.append(CuspFamily(N, d))
        elif n > 1:
            c.extend(CuspFamily(N, d, label=str(i + 1))
                     for i in range(n))
    return c


@richcmp_method
class CuspFamily(SageObject):
    r"""
    A family of elliptic curves parametrising a region of `X_0(N)`.
    """
    def __init__(self, N, width, label=None) -> None:
        r"""
        Create the cusp of width d on X_0(N) corresponding to the family
        of Tate curves `(\CC_p/q^d, \langle \zeta q\rangle)`.

        Here `\zeta` is a primitive root of unity of order `r` with
        `\mathrm{lcm}(r,d) = N`. The cusp does not store zeta, so we
        store an arbitrary label instead.

        EXAMPLES::

            sage: CuspFamily(8, 4)
            (c_{4})
            sage: CuspFamily(16, 4, '1')
            (c_{4,1})
        """
        N = ZZ(N)
        if N <= 0:
            raise ValueError("N must be positive")
        self._N = N
        self._width = width
        if N % width:
            raise ValueError("bad width")
        if n_cusps_of_width(N, width) > 1 and label is None:
            raise ValueError("there are %s > 1 cusps of width %s on X_0(%s): specify a label" % (n_cusps_of_width(N, width), width, N))
        if n_cusps_of_width(N, width) == 1 and label is not None:
            raise ValueError("there is only one cusp of width %s on X_0(%s): no need to specify a label" % (width, N))
        self.label = label

    @property
    def __tuple(self):
        """
        The defining data of this ``CuspFamily`` as tuple, used for
        comparisons.
        """
        return (self._N, self._width, self.label)

    def __richcmp__(self, other, op) -> bool:
        """
        EXAMPLES::

            sage: a = CuspFamily(16, 4, "1"); a
            (c_{4,1})
            sage: b = CuspFamily(16, 4, "2"); b
            (c_{4,2})
            sage: c = CuspFamily(8, 8); c
            (0)
            sage: a == a
            True
            sage: a == b
            False
            sage: a != b
            True
            sage: a == c
            False
            sage: a < c
            False
            sage: a > c
            True
            sage: a != "foo"
            True
        """
        if not isinstance(other, CuspFamily):
            return NotImplemented
        return richcmp(self.__tuple, other.__tuple, op)

    def __hash__(self):
        """
        EXAMPLES::

            sage: hash(CuspFamily(10, 1))  # random
            -4769758480201659164
        """
        return hash(self.__tuple)

    def width(self) -> Integer:
        r"""
        Return the width of this cusp.

        EXAMPLES::

            sage: e = CuspFamily(10, 1)
            sage: e.width()
            1
        """
        return self._width

    def level(self) -> Integer:
        r"""
        Return the level of this cusp.

        EXAMPLES::

            sage: e = CuspFamily(10, 1)
            sage: e.level()
            10
        """
        return self._N

    def sage_cusp(self):
        r"""
        Return the corresponding element of `\mathbb{P}^1(\QQ)`.

        EXAMPLES::

            sage: CuspFamily(10, 1).sage_cusp() # not implemented
            Infinity
        """
        raise NotImplementedError

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: CuspFamily(16, 4, "1")._repr_()
            '(c_{4,1})'
        """
        if self.width() == 1:
            return "(Inf)"
        if self.width() == self.level():
            return "(0)"
        return "(c_{%s%s})" % (self.width(), ((self.label and ("," + self.label)) or ""))


def qexp_eta(ps_ring, prec):
    r"""
    Return the `q`-expansion of `\eta(q) / q^{1/24}`.

    Here `\eta(q)` is Dedekind's function

    .. MATH::

        \eta(q) = q^{1/24}\prod_{n=1}^\infty (1-q^n).

    The result is an element of ``ps_ring``, with precision ``prec``.

    INPUT:

    - ``ps_ring`` -- PowerSeriesRing; a power series ring

    - ``prec`` -- integer; the number of terms to compute

    OUTPUT: an element of ``ps_ring`` which is the `q`-expansion of
    `\eta(q)/q^{1/24}` truncated to prec terms.

    ALGORITHM: We use the Euler identity

    .. MATH::

         \eta(q) = q^{1/24}( 1 + \sum_{n \ge 1} (-1)^n (q^{n(3n+1)/2} + q^{n(3n-1)/2})

    to compute the expansion.

    EXAMPLES::

        sage: from sage.modular.etaproducts import qexp_eta
        sage: qexp_eta(ZZ[['q']], 100)
        1 - q - q^2 + q^5 + q^7 - q^12 - q^15 + q^22 + q^26 - q^35 - q^40 + q^51 + q^57 - q^70 - q^77 + q^92 + O(q^100)
    """
    prec = Integer(prec)
    if not prec > 0:
        raise ValueError("prec must be a positive integer")
    v = [Integer(0)] * prec
    pm = Integer(1)
    v[0] = pm
    try:
        n = 1
        while True:
            pm = -pm
            v[n * (3 * n - 1) // 2] = pm
            v[n * (3 * n + 1) // 2] = pm
            n += 1
    except IndexError:
        pass
    return ps_ring(v, prec=prec)


def eta_poly_relations(eta_elements, degree, labels=['x1', 'x2'],
                       verbose=False):
    r"""
    Find polynomial relations between eta products.

    INPUT:

    - ``eta_elements`` -- list; a list of EtaGroupElement objects.
      Not implemented unless this list has precisely two elements. degree

    - ``degree`` -- integer; the maximal degree of polynomial to look for

    - ``labels`` -- list of strings; labels to use for the polynomial returned

    - ``verbose`` -- boolean (default: ``False``); if ``True``, prints information as
      it goes

    OUTPUT: list of polynomials which is a Groebner basis for the
    part of the ideal of relations between eta_elements which is
    generated by elements up to the given degree; or None, if no
    relations were found.

    ALGORITHM: An expression of the form
    `\sum_{0 \le i,j \le d} a_{ij} x^i y^j` is zero if and
    only if it vanishes at the cusp infinity to degree at least
    `v = d(deg(x) + deg(y))`. For all terms up to `q^v`
    in the `q`-expansion of this expression to be zero is a
    system of `v + k` linear equations in `d^2`
    coefficients, where `k` is the number of nonzero negative
    coefficients that can appear.

    Solving these equations and calculating a basis for the solution
    space gives us a set of polynomial relations, but this is generally
    far from a minimal generating set for the ideal, so we calculate a
    Groebner basis.

    As a test, we calculate five extra terms of `q`-expansion
    and check that this doesn't change the answer.

    EXAMPLES::

        sage: from sage.modular.etaproducts import eta_poly_relations
        sage: t = EtaProduct(26, {2:2,13:2,26:-2,1:-2})
        sage: u = EtaProduct(26, {2:4,13:2,26:-4,1:-2})
        sage: eta_poly_relations([t, u], 3)
        sage: eta_poly_relations([t, u], 4)
        [x1^3*x2 - 13*x1^3 - 4*x1^2*x2 - 4*x1*x2 - x2^2 + x2]

    Use ``verbose=True`` to see the details of the computation::

        sage: eta_poly_relations([t, u], 3, verbose=True)
        Trying to find a relation of degree 3
        Lowest order of a term at infinity = -12
        Highest possible degree of a term = 15
        Trying all coefficients from q^-12 to q^15 inclusive
        No polynomial relation of order 3 valid for 28 terms
        Check:
        Trying all coefficients from q^-12 to q^20 inclusive
        No polynomial relation of order 3 valid for 33 terms

    ::

        sage: eta_poly_relations([t, u], 4, verbose=True)
        Trying to find a relation of degree 4
        Lowest order of a term at infinity = -16
        Highest possible degree of a term = 20
        Trying all coefficients from q^-16 to q^20 inclusive
        Check:
        Trying all coefficients from q^-16 to q^25 inclusive
        [x1^3*x2 - 13*x1^3 - 4*x1^2*x2 - 4*x1*x2 - x2^2 + x2]
    """
    if len(eta_elements) > 2:
        raise NotImplementedError("do not know how to find relations between more than two elements")

    eta1, eta2 = eta_elements

    if verbose:
        print("Trying to find a relation of degree %s" % degree)
    inf = CuspFamily(eta1.level(), 1)
    loterm = -(min(0, eta1.order_at_cusp(inf)) + min(0, eta2.order_at_cusp(inf))) * degree
    if verbose:
        print("Lowest order of a term at infinity = %s" % -loterm)

    maxdeg = sum([eta1.degree(), eta2.degree()]) * degree
    if verbose:
        print("Highest possible degree of a term = %s" % maxdeg)
    m = loterm + maxdeg + 1
    oldgrob = _eta_relations_helper(eta1, eta2, degree, m, labels, verbose)
    if verbose:
        print("Check:")
    newgrob = _eta_relations_helper(eta1, eta2, degree, m + 5, labels, verbose)
    if oldgrob != newgrob:
        raise ArithmeticError("Check: answers different!")
    return newgrob


def _eta_relations_helper(eta1, eta2, degree, qexp_terms, labels, verbose):
    r"""
    Helper function used by eta_poly_relations. Finds a basis for the
    space of linear relations between the first qexp_terms of the
    `q`-expansions of the monomials
    `\eta_1^i * \eta_2^j` for `0 \le i,j < degree`,
    and calculates a Groebner basis for the ideal generated by these
    relations.

    Liable to return meaningless results if qexp_terms isn't at least
    `1 + d*(m_1,m_2)` where

    .. MATH::

       m_i = min(0, {\text degree of the pole of $\eta_i$ at $\infty$})

    as then 1 will be in the ideal.

    EXAMPLES::

        sage: from sage.modular.etaproducts import _eta_relations_helper
        sage: r,s = EtaGroup(4).basis()
        sage: _eta_relations_helper(r,s,4,100,['a','b'],False)
        [a + 1/16*b - 1/16]
        sage: _eta_relations_helper(EtaProduct(26, {2:2,13:2,26:-2,1:-2}),EtaProduct(26, {2:4,13:2,26:-4,1:-2}),3,12,['a','b'],False) # not enough terms, will return rubbish
        [1]
    """
    indices = [(i, j) for j in range(degree) for i in range(degree)]
    inf = CuspFamily(eta1.level(), 1)

    pole_at_infinity = -(min(0, eta1.order_at_cusp(inf)) + min(0, eta2.order_at_cusp(inf))) * degree
    if verbose:
        print("Trying all coefficients from q^%s to q^%s inclusive" % (-pole_at_infinity, -pole_at_infinity + qexp_terms - 1))

    rows: list[list] = [[] for _ in range(qexp_terms)]
    for i in indices:
        func = (eta1**i[0] * eta2**i[1]).qexp(qexp_terms)
        for j in range(qexp_terms):
            rows[j].append(func[j - pole_at_infinity])
    M = matrix(rows)
    V = M.right_kernel()
    if V.dimension() == 0:
        if verbose:
            print("No polynomial relation of order %s valid for %s terms" % (degree, qexp_terms))
        return None
    if V.dimension() >= 1:
        R = PolynomialRing(QQ, 2, labels)
        x, y = R.gens()
        relations = [sum([c[v] * x**indices[v][0] * y**indices[v][1]
                          for v in range(len(indices))])
                     for c in V.basis()]
        id = R.ideal(relations)
        return id.groebner_basis()




from sage.misc.lazy_import import lazy_import

lazy_import("sage.coding.code_constructions", ["permutation_action",
                                               "walsh_matrix"])

lazy_import("sage.coding.linear_code", "LinearCode")

# Functions removed from the global namespace

lazy_import('sage.coding', 'codes_catalog', 'codes')
lazy_import('sage.coding', 'channels_catalog', 'channels')

del lazy_import



from sage.libs.ntl.ZZ_pEX cimport ZZ_pEX_c
from sage.libs.ntl.ntl_ZZ_pEContext cimport ZZ_pEContext_ptrs
from sage.rings.integer cimport Integer

ctypedef ZZ_pEX_c celement
ctypedef ZZ_pEContext_ptrs *cparent

include "polynomial_template_header.pxi"

cdef class Polynomial_ZZ_pEX(Polynomial_template):
    cdef _powmod_bigexp(Polynomial_ZZ_pEX self, Integer exp, Polynomial_ZZ_pEX modulus)



r"""
Morphisms defined by a matrix

A matrix morphism is a morphism that is defined by multiplication
by a matrix. Elements of domain must either have a method
``vector()`` that returns a vector that the defining
matrix can hit from the left, or be coercible into vector space of
appropriate dimension.

EXAMPLES::

    sage: from sage.modules.matrix_morphism import MatrixMorphism
    sage: V = QQ^3
    sage: T = End(V)
    sage: M = MatrixSpace(QQ,3)
    sage: I = M.identity_matrix()
    sage: m = MatrixMorphism(T, I); m
    Morphism defined by the matrix
    [1 0 0]
    [0 1 0]
    [0 0 1]
    sage: m.charpoly('x')                                                               # needs sage.libs.pari
    x^3 - 3*x^2 + 3*x - 1
    sage: m.base_ring()
    Rational Field
    sage: m.det()
    1
    sage: m.fcp('x')                                                                    # needs sage.libs.pari
    (x - 1)^3
    sage: m.matrix()
    [1 0 0]
    [0 1 0]
    [0 0 1]
    sage: m.rank()
    3
    sage: m.trace()
    3

AUTHOR:

- William Stein: initial versions

- David Joyner (2005-12-17): added examples

- William Stein (2005-01-07): added __reduce__

- Craig Citro (2008-03-18): refactored MatrixMorphism class

- Rob Beezer (2011-07-15): additional methods, bug fixes, documentation
"""

import sage.categories.morphism
import sage.categories.homset
from sage.categories.finite_dimensional_modules_with_basis import FiniteDimensionalModulesWithBasis
from sage.structure.sequence import Sequence
from sage.structure.element import parent
from sage.structure.richcmp import richcmp, op_NE, op_EQ


class MatrixMorphism_abstract(sage.categories.morphism.Morphism):

    # Copy in methods that delegate to self.matrix.
    # This is needed because MatrixMorphism_abstract is subclassed
    # for use with parents that are merely set up as additive abelian groups,
    # but not as ZZ-modules; see sage.modular.abvar.

    characteristic_polynomial = charpoly = FiniteDimensionalModulesWithBasis.Homsets.Endset.ElementMethods.characteristic_polynomial
    det = determinant = FiniteDimensionalModulesWithBasis.Homsets.Endset.ElementMethods.determinant
    fcp = FiniteDimensionalModulesWithBasis.Homsets.Endset.ElementMethods.fcp
    trace = FiniteDimensionalModulesWithBasis.Homsets.Endset.ElementMethods.trace

    def __init__(self, parent, side='left'):
        """
        INPUT:

        - ``parent`` -- a homspace

        - ``A`` -- matrix

        EXAMPLES::

            sage: from sage.modules.matrix_morphism import MatrixMorphism
            sage: T = End(ZZ^3)
            sage: M = MatrixSpace(ZZ,3)
            sage: I = M.identity_matrix()
            sage: A = MatrixMorphism(T, I)
            sage: loads(A.dumps()) == A
            True
        """
        if not isinstance(parent, sage.categories.homset.Homset):
            raise TypeError("parent must be a Hom space")
        if side not in ["left", "right"]:
            raise ValueError("the argument side must be either 'left' or 'right'")
        self._side = side
        sage.categories.morphism.Morphism.__init__(self, parent)

    def _richcmp_(self, other, op):
        """
        Rich comparison of morphisms.

        EXAMPLES::

            sage: V = ZZ^2
            sage: phi = V.hom([3*V.0, 2*V.1])
            sage: psi = V.hom([5*V.0, 5*V.1])
            sage: id = V.hom([V.0, V.1])
            sage: phi == phi
            True
            sage: phi == psi
            False
            sage: psi == End(V)(5)
            True
            sage: psi == 5 * id
            True
            sage: psi == 5  # no coercion
            False
            sage: id == End(V).identity()
            True
        """
        if not isinstance(other, MatrixMorphism) or op not in (op_EQ, op_NE):
            # Generic comparison
            return sage.categories.morphism.Morphism._richcmp_(self, other, op)
        return richcmp(self.matrix(), other.matrix(), op)

    def _call_(self, x):
        """
        Evaluate this matrix morphism at an element of the domain.

        .. NOTE::

            Coercion is done in the generic :meth:`__call__` method,
            which calls this method.

        EXAMPLES::

            sage: V = QQ^3; W = QQ^2
            sage: H = Hom(V, W); H
            Set of Morphisms (Linear Transformations) from
            Vector space of dimension 3 over Rational Field to
            Vector space of dimension 2 over Rational Field
            sage: phi = H(matrix(QQ, 3, 2, range(6))); phi
            Vector space morphism represented by the matrix:
            [0 1]
            [2 3]
            [4 5]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 2 over Rational Field
            sage: phi(V.0)
            (0, 1)
            sage: phi(V([1, 2, 3]))
            (16, 22)

        Last, we have a situation where coercion occurs::

            sage: U = V.span([[3,2,1]])
            sage: U.0
            (1, 2/3, 1/3)
            sage: phi(2*U.0)
            (16/3, 28/3)

        TESTS::

            sage: V = QQ^3; W = span([[1,2,3],[-1,2,5/3]], QQ)
            sage: phi = V.hom(matrix(QQ,3,[1..9]))

        We compute the image of some elements::

            sage: phi(V.0)    #indirect doctest
            (1, 2, 3)
            sage: phi(V.1)
            (4, 5, 6)
            sage: phi(V.0  - 1/4*V.1)
            (0, 3/4, 3/2)

        We restrict ``phi`` to ``W`` and compute the image of an element::

            sage: psi = phi.restrict_domain(W)
            sage: psi(W.0) == phi(W.0)
            True
            sage: psi(W.1) == phi(W.1)
            True
        """
        try:
            if parent(x) is not self.domain():
                x = self.domain()(x)
        except TypeError:
            raise TypeError("%s must be coercible into %s" % (x, self.domain()))
        if self.domain().is_ambient():
            x = x.element()
        else:
            x = self.domain().coordinate_vector(x)
        C = self.codomain()
        if self.side() == "left":
            v = x.change_ring(C.base_ring()) * self.matrix()
        else:
            v = self.matrix() * x.change_ring(C.base_ring())
        if not C.is_ambient():
            v = C.linear_combination_of_basis(v)
        # The call method of parents uses (coercion) morphisms.
        # Hence, in order to avoid recursion, we call the element
        # constructor directly; after all, we already know the
        # coordinates.
        return C._element_constructor_(v)

    def _call_with_args(self, x, args=(), kwds={}):
        """
        Like :meth:`_call_`, but takes optional and keyword arguments.

        EXAMPLES::

            sage: # needs sage.rings.real_mpfr sage.symbolic
            sage: V = RR^2
            sage: f = V.hom(V.gens())
            sage: f._matrix *= I         # f is now invalid
            sage: f((1, 0))
            Traceback (most recent call last):
            ...
            TypeError: Unable to coerce entries (=[1.00000000000000*I, 0.000000000000000])
            to coefficients in Real Field with 53 bits of precision
            sage: f((1, 0), coerce=False)
            (1.00000000000000*I, 0.000000000000000)
        """
        if self.domain().is_ambient():
            x = x.element()
        else:
            x = self.domain().coordinate_vector(x)
        C = self.codomain()
        v = x.change_ring(C.base_ring()) * self.matrix()
        if not C.is_ambient():
            v = C.linear_combination_of_basis(v)
        # The call method of parents uses (coercion) morphisms.
        # Hence, in order to avoid recursion, we call the element
        # constructor directly; after all, we already know the
        # coordinates.
        return C._element_constructor_(v, *args, **kwds)

    def __invert__(self):
        """
        Invert this matrix morphism.

        EXAMPLES::

            sage: V = QQ^2; phi = V.hom([3*V.0, 2*V.1])
            sage: phi^(-1)
            Vector space morphism represented by the matrix:
            [1/3   0]
            [  0 1/2]
            Domain: Vector space of dimension 2 over Rational Field
            Codomain: Vector space of dimension 2 over Rational Field

        Check that a certain non-invertible morphism isn't invertible::

            sage: V = ZZ^2; phi = V.hom([3*V.0, 2*V.1])
            sage: phi^(-1)
            Traceback (most recent call last):
            ...
            ZeroDivisionError: matrix morphism not invertible
        """
        try:
            B = ~(self.matrix())
        except ZeroDivisionError:
            raise ZeroDivisionError("matrix morphism not invertible")
        try:
            return self.parent().reversed()(B, side=self.side())
        except TypeError:
            raise ZeroDivisionError("matrix morphism not invertible")

    def side(self):
        """
        Return the side of vectors acted on, relative to the matrix.

        EXAMPLES::

            sage: m = matrix(2, [1, 1, 0, 1])
            sage: V = ZZ^2
            sage: h1 = V.hom(m); h2 = V.hom(m, side='right')
            sage: h1.side()
            'left'
            sage: h1([1, 0])
            (1, 1)
            sage: h2.side()
            'right'
            sage: h2([1, 0])
            (1, 0)
        """
        return self._side

    def side_switch(self):
        """
        Return the same morphism, acting on vectors on the opposite side.

        EXAMPLES::

            sage: m = matrix(2, [1,1,0,1]); m
            [1 1]
            [0 1]
            sage: V = ZZ^2
            sage: h = V.hom(m); h.side()
            'left'
            sage: h2 = h.side_switch(); h2
            Free module morphism defined as left-multiplication by the matrix
            [1 0]
            [1 1]
            Domain: Ambient free module of rank 2 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 2 over the principal ideal domain Integer Ring
            sage: h2.side()
            'right'
            sage: h2.side_switch().matrix()
            [1 1]
            [0 1]
        """
        side = "left" if self.side() == "right" else "right"
        return self.parent()(self.matrix().transpose(), side=side)

    def inverse(self):
        r"""
        Return the inverse of this matrix morphism, if the inverse exists.

        This raises a :exc:`ZeroDivisionError` if the inverse does not exist.

        EXAMPLES:

        An invertible morphism created as a restriction of
        a non-invertible morphism, and which has an unequal
        domain and codomain.  ::

            sage: V = QQ^4
            sage: W = QQ^3
            sage: m = matrix(QQ, [[2, 0, 3], [-6, 1, 4], [1, 2, -4], [1, 0, 1]])
            sage: phi = V.hom(m, W)
            sage: rho = phi.restrict_domain(V.span([V.0, V.3]))
            sage: zeta = rho.restrict_codomain(W.span([W.0, W.2]))
            sage: x = vector(QQ, [2, 0, 0, -7])
            sage: y = zeta(x); y
            (-3, 0, -1)
            sage: inv = zeta.inverse(); inv
            Vector space morphism represented by the matrix:
            [-1  3]
            [ 1 -2]
            Domain: Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 0 1]
            Codomain: Vector space of degree 4 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0 0]
            [0 0 0 1]
            sage: inv(y) == x
            True

        An example of an invertible morphism between modules,
        (rather than between vector spaces).  ::

            sage: M = ZZ^4
            sage: p = matrix(ZZ, [[ 0, -1,  1, -2],
            ....:                 [ 1, -3,  2, -3],
            ....:                 [ 0,  4, -3,  4],
            ....:                 [-2,  8, -4,  3]])
            sage: phi = M.hom(p, M)
            sage: x = vector(ZZ, [1, -3, 5, -2])
            sage: y = phi(x); y
            (1, 12, -12, 21)
            sage: rho = phi.inverse(); rho
            Free module morphism defined by the matrix
            [ -5   3  -1   1]
            [ -9   4  -3   2]
            [-20   8  -7   4]
            [ -6   2  -2   1]
            Domain: Ambient free module of rank 4 over the principal ideal domain ...
            Codomain: Ambient free module of rank 4 over the principal ideal domain ...
            sage: rho(y) == x
            True

        A non-invertible morphism, despite having an appropriate
        domain and codomain.  ::

            sage: V = QQ^2
            sage: m = matrix(QQ, [[1, 2], [20, 40]])
            sage: phi = V.hom(m, V)
            sage: phi.is_bijective()
            False
            sage: phi.inverse()
            Traceback (most recent call last):
            ...
            ZeroDivisionError: matrix morphism not invertible

        The matrix representation of this morphism is invertible
        over the rationals, but not over the integers, thus the
        morphism is not invertible as a map between modules.
        It is easy to notice from the definition that every
        vector of the image will have a second entry that
        is an even integer.  ::

            sage: V = ZZ^2
            sage: q = matrix(ZZ, [[1, 2], [3, 4]])
            sage: phi = V.hom(q, V)
            sage: phi.matrix().change_ring(QQ).inverse()
            [  -2    1]
            [ 3/2 -1/2]
            sage: phi.is_bijective()
            False
            sage: phi.image()
            Free module of degree 2 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1 0]
            [0 2]
            sage: phi.lift(vector(ZZ, [1, 1]))
            Traceback (most recent call last):
            ...
            ValueError: element is not in the image
            sage: phi.inverse()
            Traceback (most recent call last):
            ...
            ZeroDivisionError: matrix morphism not invertible

        The unary invert operator (~, tilde, "wiggle") is synonymous
        with the ``inverse()`` method (and a lot easier to type).  ::

            sage: V = QQ^2
            sage: r = matrix(QQ, [[4, 3], [-2, 5]])
            sage: phi = V.hom(r, V)
            sage: rho = phi.inverse()
            sage: zeta = ~phi
            sage: rho.is_equal_function(zeta)
            True

        TESTS::

            sage: V = QQ^2
            sage: W = QQ^3
            sage: U = W.span([W.0, W.1])
            sage: m = matrix(QQ, [[2, 1], [3, 4]])
            sage: phi = V.hom(m, U)
            sage: inv = phi.inverse()
            sage: (inv*phi).is_identity()
            True
            sage: (phi*inv).is_identity()
            True
        """
        return ~self

    def __rmul__(self, left):
        """
        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([V.0+V.1, 2*V.1])
            sage: 2*phi
            Free module morphism defined by the matrix
            [2 2]
            [0 4]...
            sage: phi*2
            Free module morphism defined by the matrix
            [2 2]
            [0 4]...
        """
        R = self.base_ring()
        return self.parent()(R(left) * self.matrix(), side=self.side())

    def __mul__(self, right):
        r"""
        Composition of morphisms, denoted by \*.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([V.0+V.1, 2*V.1])
            sage: phi*phi
            Free module morphism defined by the matrix
            [1 3]
            [0 4]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...

            sage: V = QQ^3
            sage: E = V.endomorphism_ring()
            sage: phi = E(Matrix(QQ,3,range(9))) ; phi
            Vector space morphism represented by the matrix:
            [0 1 2]
            [3 4 5]
            [6 7 8]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 3 over Rational Field
            sage: phi*phi
            Vector space morphism represented by the matrix:
            [ 15  18  21]
            [ 42  54  66]
            [ 69  90 111]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 3 over Rational Field
            sage: phi.matrix()**2
            [ 15  18  21]
            [ 42  54  66]
            [ 69  90 111]

        ::

            sage: W = QQ**4
            sage: E_VW = V.Hom(W)
            sage: psi = E_VW(Matrix(QQ,3,4,range(12))) ; psi
            Vector space morphism represented by the matrix:
            [ 0  1  2  3]
            [ 4  5  6  7]
            [ 8  9 10 11]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 4 over Rational Field
            sage: psi*phi
            Vector space morphism represented by the matrix:
            [ 20  23  26  29]
            [ 56  68  80  92]
            [ 92 113 134 155]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 4 over Rational Field
            sage: phi*psi
            Traceback (most recent call last):
            ...
            TypeError: Incompatible composition of morphisms: domain of left morphism must be codomain of right.
            sage: phi.matrix()*psi.matrix()
            [ 20  23  26  29]
            [ 56  68  80  92]
            [ 92 113 134 155]

        Composite maps can be formed with matrix morphisms::

            sage: # needs sage.rings.number_field
            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 + 23)
            sage: V, VtoK, KtoV = K.vector_space()
            sage: f = V.hom([V.0 - V.1, V.0 + V.1])*KtoV; f
            Composite map:
            From: Number Field in a with defining polynomial x^2 + 23
            To:   Vector space of dimension 2 over Rational Field
            Defn:   Isomorphism map:
                    From: Number Field in a with defining polynomial x^2 + 23
                    To:   Vector space of dimension 2 over Rational Field
                    then
                    Vector space morphism represented by the matrix:
                    [ 1 -1]
                    [ 1  1]
                    Domain: Vector space of dimension 2 over Rational Field
                    Codomain: Vector space of dimension 2 over Rational Field
            sage: f(a)
            (1, 1)
            sage: V.hom([V.0 - V.1, V.0 + V.1], side='right')*KtoV
            Composite map:
              From: Number Field in a with defining polynomial x^2 + 23
              To:   Vector space of dimension 2 over Rational Field
              Defn:   Isomorphism map:
                      From: Number Field in a with defining polynomial x^2 + 23
                      To:   Vector space of dimension 2 over Rational Field
                    then
                      Vector space morphism represented as left-multiplication by the matrix:
                    [ 1  1]
                    [-1  1]
                    Domain: Vector space of dimension 2 over Rational Field
                    Codomain: Vector space of dimension 2 over Rational Field


        We can test interaction between morphisms with different ``side``::

            sage: V = ZZ^2
            sage: m = matrix(2, [1,1,0,1])
            sage: hl = V.hom(m)
            sage: hr = V.hom(m, side='right')
            sage: hl * hl
            Free module morphism defined by the matrix
            [1 2]
            [0 1]...
            sage: hl * hr
            Free module morphism defined by the matrix
            [1 1]
            [1 2]...
            sage: hl * hl.side_switch()
            Free module morphism defined by the matrix
            [1 2]
            [0 1]...
            sage: hr * hl
            Free module morphism defined by the matrix
            [2 1]
            [1 1]...
            sage: hl * hl
            Free module morphism defined by the matrix
            [1 2]
            [0 1]...
            sage: hr / hl
            Free module morphism defined by the matrix
            [ 0 -1]
            [ 1  1]...
            sage: hr / hr.side_switch()
            Free module morphism defined by the matrix
            [1 0]
            [0 1]...
            sage: hl / hl
            Free module morphism defined by the matrix
            [1 0]
            [0 1]...
            sage: hr / hr
            Free module morphism defined as left-multiplication by the matrix
            [1 0]
            [0 1]...

        .. WARNING::

            Matrix morphisms can be defined by either left or right-multiplication.
            The composite morphism always applies the morphism on the right of \* first.
            The matrix of the composite morphism of two morphisms given by
            right-multiplication is not the morphism given by the product of their
            respective matrices.
            If the two morphisms act on different sides, then the side of the resulting
            morphism is the default one.
        """
        if not isinstance(right, MatrixMorphism):
            if isinstance(right, (sage.categories.morphism.Morphism, sage.categories.map.Map)):
                return sage.categories.map.Map.__mul__(self, right)
            R = self.base_ring()
            return self.parent()(self.matrix() * R(right))
        H = right.domain().Hom(self.codomain())
        if self.domain() != right.codomain():
            raise TypeError("Incompatible composition of morphisms: domain of left morphism must be codomain of right.")
        if self.side() == "left":
            if right.side() == "left":
                return H(right.matrix() * self.matrix(), side=self.side())
            return H(right.matrix().transpose() * self.matrix(), side=self.side())
        if right.side() == "right":
            return H(self.matrix() * right.matrix(), side=self.side())
        return H(right.matrix() * self.matrix().transpose(), side='left')

    def __add__(self, right):
        """
        Sum of morphisms, denoted by +.

        EXAMPLES::

            sage: phi = (ZZ**2).endomorphism_ring()(Matrix(ZZ,2,[2..5])) ; phi
            Free module morphism defined by the matrix
            [2 3]
            [4 5]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...
            sage: phi + 3
            Free module morphism defined by the matrix
            [5 3]
            [4 8]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...
            sage: phi + phi
            Free module morphism defined by the matrix
            [ 4  6]
            [ 8 10]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...
            sage: psi = (ZZ**3).endomorphism_ring()(Matrix(ZZ,3,[22..30])) ; psi
            Free module morphism defined by the matrix
            [22 23 24]
            [25 26 27]
            [28 29 30]
            Domain: Ambient free module of rank 3 over the principal ideal domain ...
            Codomain: Ambient free module of rank 3 over the principal ideal domain ...
            sage: phi + psi
            Traceback (most recent call last):
            ...
            ValueError: inconsistent number of rows: should be 2 but got 3

        ::


            sage: V = ZZ^2
            sage: m = matrix(2, [1,1,0,1])
            sage: hl = V.hom(m)
            sage: hr = V.hom(m, side='right')
            sage: hl + hl
            Free module morphism defined by the matrix
            [2 2]
            [0 2]...
            sage: hr + hr
            Free module morphism defined as left-multiplication by the matrix
            [2 2]
            [0 2]...
            sage: hr + hl
            Free module morphism defined by the matrix
            [2 1]
            [1 2]...
            sage: hl + hr
            Free module morphism defined by the matrix
            [2 1]
            [1 2]...

        .. WARNING::

            If the two morphisms do not share the same ``side`` attribute, then
            the resulting morphism will be defined with the default value.
        """
        # TODO: move over to any coercion model!
        if not isinstance(right, MatrixMorphism):
            R = self.base_ring()
            return self.parent()(self.matrix() + R(right))
        if not right.parent() == self.parent():
            right = self.parent()(right, side=right.side())
        if self.side() == "left":
            if right.side() == "left":
                return self.parent()(self.matrix() + right.matrix(), side=self.side())
            if right.side() == "right":
                return self.parent()(self.matrix() + right.matrix().transpose(), side='left')
        if self.side() == "right":
            if right.side() == "right":
                return self.parent()(self.matrix() + right.matrix(), side=self.side())
            if right.side() == "left":
                return self.parent()(self.matrix().transpose() + right.matrix(), side='left')

    def __neg__(self):
        """
        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([V.0+V.1, 2*V.1])
            sage: -phi
            Free module morphism defined by the matrix
            [-1 -1]
            [ 0 -2]...
            sage: phi2 = phi.side_switch(); -phi2
            Free module morphism defined as left-multiplication by the matrix
            [-1  0]
            [-1 -2]...
        """
        return self.parent()(-self.matrix(), side=self.side())

    def __sub__(self, other):
        """
        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([V.0+V.1, 2*V.1])
            sage: phi - phi
            Free module morphism defined by the matrix
            [0 0]
            [0 0]...

        ::

            sage: V = ZZ^2
            sage: m = matrix(2, [1,1,0,1])
            sage: hl = V.hom(m)
            sage: hr = V.hom(m, side='right')
            sage: hl - hr
            Free module morphism defined by the matrix
            [ 0  1]
            [-1  0]...
            sage: hl - hl
            Free module morphism defined by the matrix
            [0 0]
            [0 0]...
            sage: hr - hr
            Free module morphism defined as left-multiplication by the matrix
            [0 0]
            [0 0]...
            sage: hr-hl
            Free module morphism defined by the matrix
            [ 0 -1]
            [ 1  0]...

        .. WARNING::

            If the two morphisms do not share the same ``side`` attribute, then
            the resulting morphism will be defined with the default value.
        """
        # TODO: move over to any coercion model!
        if not isinstance(other, MatrixMorphism):
            R = self.base_ring()
            return self.parent()(self.matrix() - R(other), side=self.side())
        if not other.parent() == self.parent():
            other = self.parent()(other, side=other.side())
        if self.side() == "left":
            if other.side() == "left":
                return self.parent()(self.matrix() - other.matrix(), side=self.side())
            if other.side() == "right":
                return self.parent()(self.matrix() - other.matrix().transpose(), side='left')
        if self.side() == "right":
            if other.side() == "right":
                return self.parent()(self.matrix() - other.matrix(), side=self.side())
            if other.side() == "left":
                return self.parent()(self.matrix().transpose() - other.matrix(), side='left')

    def base_ring(self):
        """
        Return the base ring of ``self``, that is, the ring over which ``self``
        is given by a matrix.

        EXAMPLES::

            sage: sage.modules.matrix_morphism.MatrixMorphism((ZZ**2).endomorphism_ring(), Matrix(ZZ,2,[3..6])).base_ring()
            Integer Ring
        """
        return self.domain().base_ring()

    def decomposition(self, *args, **kwds):
        """
        Return decomposition of this endomorphism, i.e., sequence of
        subspaces obtained by finding invariant subspaces of ``self``.

        See the documentation for ``self.matrix().decomposition`` for more
        details.  All inputs to this function are passed onto the
        matrix one.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([V.0+V.1, 2*V.1])
            sage: phi.decomposition()                                                   # needs sage.libs.pari
            [Free module of degree 2 and rank 1 over Integer Ring
             Echelon basis matrix:
             [0 1],
             Free module of degree 2 and rank 1 over Integer Ring
             Echelon basis matrix:
             [ 1 -1]]
            sage: phi2 = V.hom(phi.matrix(), side='right')
            sage: phi2.decomposition()                                                  # needs sage.libs.pari
            [Free module of degree 2 and rank 1 over Integer Ring
             Echelon basis matrix:
             [1 1],
             Free module of degree 2 and rank 1 over Integer Ring
             Echelon basis matrix:
             [1 0]]
        """
        if not self.is_endomorphism():
            raise ArithmeticError("matrix morphism must be an endomorphism")
        D = self.domain()
        if self.side() == "left":
            E = self.matrix().decomposition(*args, **kwds)
        else:
            E = self.matrix().transpose().decomposition(*args, **kwds)
        if D.is_ambient():
            return Sequence([D.submodule(V, check=False) for V, _ in E],
                            cr=True, check=False)
        B = D.basis_matrix()
        R = D.base_ring()
        return Sequence([D.submodule((V.basis_matrix() * B).row_module(R),
                                     check=False) for V, _ in E],
                        cr=True, check=False)

    def kernel(self):
        """
        Compute the kernel of this morphism.

        EXAMPLES::

            sage: V = VectorSpace(QQ, 3)
            sage: id = V.Hom(V)(identity_matrix(QQ,3))
            sage: null = V.Hom(V)(0*identity_matrix(QQ,3))
            sage: id.kernel()
            Vector space of degree 3 and dimension 0 over Rational Field
            Basis matrix:
            []
            sage: phi = V.Hom(V)(matrix(QQ, 3, range(9)))
            sage: phi.kernel()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [ 1 -2  1]
            sage: hom(CC^2, CC^2, matrix(CC, [[1,0], [0,1]])).kernel()
            Vector space of degree 2 and dimension 0 over Complex Field with 53 bits of precision
            Basis matrix:
            []
            sage: m = matrix(3, [1, 0, 0, 1, 0, 0, 0, 0, 1]); m
            [1 0 0]
            [1 0 0]
            [0 0 1]
            sage: f1 = V.hom(m)
            sage: f2 = V.hom(m, side='right')
            sage: f1.kernel()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [ 1 -1  0]
            sage: f2.kernel()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [0 1 0]
        """
        if self.side() == "left":
            V = self.matrix().left_kernel()
        else:
            V = self.matrix().right_kernel()
        D = self.domain()
        if not D.is_ambient():
            # Transform V to ambient space
            # This is a matrix multiply:  we take the linear combinations of the basis for
            # D given by the elements of the basis for V.
            B = V.basis_matrix() * D.basis_matrix()
            V = B.row_module(D.base_ring())
        return self.domain().submodule(V, check=False)

    def image(self):
        """
        Compute the image of this morphism.

        EXAMPLES::

            sage: V = VectorSpace(QQ, 3)
            sage: phi = V.Hom(V)(matrix(QQ, 3, range(9)))
            sage: phi.image()
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [ 1  0 -1]
            [ 0  1  2]
            sage: hom(GF(7)^3, GF(7)^2, zero_matrix(GF(7), 3, 2)).image()
            Vector space of degree 2 and dimension 0 over Finite Field of size 7
            Basis matrix:
            []
            sage: m = matrix(3, [1, 0, 0, 1, 0, 0, 0, 0, 1]); m
            [1 0 0]
            [1 0 0]
            [0 0 1]
            sage: f1 = V.hom(m)
            sage: f2 = V.hom(m, side='right')
            sage: f1.image()
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 0 1]
            sage: f2.image()
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 1 0]
            [0 0 1]


        Compute the image of the identity map on a ZZ-submodule::

            sage: V = (ZZ^2).span([[1,2], [3,4]])
            sage: phi = V.Hom(V)(identity_matrix(ZZ,2))
            sage: phi(V.0) == V.0
            True
            sage: phi(V.1) == V.1
            True
            sage: phi.image()
            Free module of degree 2 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1 0]
            [0 2]
            sage: phi.image() == V
            True
        """
        if self.side() == 'left':
            V = self.matrix().row_space()
        else:
            V = self.matrix().column_space()
        C = self.codomain()
        if not C.is_ambient():
            # Transform V to ambient space
            # This is a matrix multiply:  we take the linear combinations of the basis for
            # D given by the elements of the basis for V.
            B = V.basis_matrix() * C.basis_matrix()
            V = B.row_module(self.domain().base_ring())
        return self.codomain().submodule(V, check=False)

    def matrix(self):
        """
        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom(V.basis())
            sage: phi.matrix()
            [1 0]
            [0 1]
            sage: sage.modules.matrix_morphism.MatrixMorphism_abstract.matrix(phi)
            Traceback (most recent call last):
            ...
            NotImplementedError: this method must be overridden in the extension class
        """
        raise NotImplementedError("this method must be overridden in the extension class")

    def _matrix_(self):
        """
        EXAMPLES:

        Check that this works with the :func:`matrix` function
        (:issue:`16844`)::

            sage: H = Hom(ZZ^2, ZZ^3)
            sage: x = H.an_element()
            sage: matrix(x)
            [0 0 0]
            [0 0 0]

        TESTS:

        ``matrix(x)`` is immutable::

            sage: H = Hom(QQ^3, QQ^2)
            sage: phi = H(matrix(QQ, 3, 2, list(reversed(range(6))))); phi
            Vector space morphism represented by the matrix:
            [5 4]
            [3 2]
            [1 0]
            Domain: Vector space of dimension 3 over Rational Field
            Codomain: Vector space of dimension 2 over Rational Field
            sage: A = phi.matrix()
            sage: A[1, 1] = 19
            Traceback (most recent call last):
            ...
            ValueError: matrix is immutable; please change a copy instead (i.e., use copy(M) to change a copy of M).
        """
        return self.matrix()

    def rank(self):
        r"""
        Return the rank of the matrix representing this morphism.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom(V.basis())
            sage: phi.rank()
            2
            sage: V = ZZ^2; phi = V.hom([V.0, V.0])
            sage: phi.rank()
            1
        """
        return self.matrix().rank()

    def nullity(self):
        r"""
        Return the nullity of the matrix representing this morphism, which is the
        dimension of its kernel.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom(V.basis())
            sage: phi.nullity()
            0
            sage: V = ZZ^2; phi = V.hom([V.0, V.0])
            sage: phi.nullity()
            1

        ::

            sage: m = matrix(2, [1, 2])
            sage: V = ZZ^2
            sage: h1 = V.hom(m)
            sage: h1.nullity()
            1
            sage: W = ZZ^1
            sage: h2 = W.hom(m, side='right')
            sage: h2.nullity()
            0
        """
        if self.side() == "left":
            return self._matrix.left_nullity()
        return self._matrix.right_nullity()

    def is_bijective(self) -> bool:
        r"""
        Tell whether ``self`` is bijective.

        EXAMPLES:

        Two morphisms that are obviously not bijective, simply on
        considerations of the dimensions.  However, each fullfills
        half of the requirements to be a bijection.  ::

            sage: V1 = QQ^2
            sage: V2 = QQ^3
            sage: m = matrix(QQ, [[1, 2, 3], [4, 5, 6]])
            sage: phi = V1.hom(m, V2)
            sage: phi.is_injective()
            True
            sage: phi.is_bijective()
            False
            sage: rho = V2.hom(m.transpose(), V1)
            sage: rho.is_surjective()
            True
            sage: rho.is_bijective()
            False

        We construct a simple bijection between two one-dimensional
        vector spaces.  ::

            sage: V1 = QQ^3
            sage: V2 = QQ^2
            sage: phi = V1.hom(matrix(QQ, [[1, 2], [3, 4], [5, 6]]), V2)
            sage: x = vector(QQ, [1, -1, 4])
            sage: y = phi(x); y
            (18, 22)
            sage: rho = phi.restrict_domain(V1.span([x]))
            sage: zeta = rho.restrict_codomain(V2.span([y]))
            sage: zeta.is_bijective()
            True

        AUTHOR:

        - Rob Beezer (2011-06-28)
        """
        return self.is_injective() and self.is_surjective()

    def is_identity(self) -> bool:
        r"""
        Determine if this morphism is an identity function or not.

        EXAMPLES:

        A homomorphism that cannot possibly be the identity
        due to an unequal domain and codomain.  ::

            sage: V = QQ^3
            sage: W = QQ^2
            sage: m = matrix(QQ, [[1, 2], [3, 4], [5, 6]])
            sage: phi = V.hom(m, W)
            sage: phi.is_identity()
            False

        A bijection, but not the identity. ::

            sage: V = QQ^3
            sage: n = matrix(QQ, [[3, 1, -8], [5, -4, 6], [1, 1, -5]])
            sage: phi = V.hom(n, V)
            sage: phi.is_bijective()
            True
            sage: phi.is_identity()
            False

        A restriction that is the identity.  ::

            sage: V = QQ^3
            sage: p = matrix(QQ, [[1, 0, 0], [5, 8, 3], [0, 0, 1]])
            sage: phi = V.hom(p, V)
            sage: rho = phi.restrict(V.span([V.0, V.2]))
            sage: rho.is_identity()
            True

        An identity linear transformation that is defined with a
        domain and codomain with wildly different bases, so that the
        matrix representation is not simply the identity matrix. ::

            sage: A = matrix(QQ, [[1, 1, 0], [2, 3, -4], [2, 4, -7]])
            sage: B = matrix(QQ, [[2, 7, -2], [-1, -3, 1], [-1, -6, 2]])
            sage: U = (QQ^3).subspace_with_basis(A.rows())
            sage: V = (QQ^3).subspace_with_basis(B.rows())
            sage: H = Hom(U, V)
            sage: id = lambda x: x
            sage: phi = H(id)
            sage: phi([203, -179, 34])
            (203, -179, 34)
            sage: phi.matrix()
            [  1   0   1]
            [ -9 -18  -2]
            [-17 -31  -5]
            sage: phi.is_identity()
            True

        TESTS::

            sage: V = QQ^10
            sage: H = Hom(V, V)
            sage: id = H.identity()
            sage: id.is_identity()
            True

        AUTHOR:

        - Rob Beezer (2011-06-28)
        """
        if self.domain() != self.codomain():
            return False
        # testing for the identity matrix will only work for
        #   endomorphisms which have the same basis for domain and codomain
        #   so we test equality on a basis, which is sufficient
        return all(self(u) == u for u in self.domain().basis())

    def is_zero(self) -> bool:
        r"""
        Determine if this morphism is a zero function or not.

        EXAMPLES:

        A zero morphism created from a function.  ::

            sage: V = ZZ^5
            sage: W = ZZ^3
            sage: z = lambda x: zero_vector(ZZ, 3)
            sage: phi = V.hom(z, W)
            sage: phi.is_zero()
            True

        An image list that just barely makes a nonzero morphism.  ::

            sage: V = ZZ^4
            sage: W = ZZ^6
            sage: z = zero_vector(ZZ, 6)
            sage: images = [z, z, W.5, z]
            sage: phi = V.hom(images, W)
            sage: phi.is_zero()
            False

        TESTS::

            sage: V = QQ^10
            sage: W = QQ^3
            sage: H = Hom(V, W)
            sage: rho = H.zero()
            sage: rho.is_zero()
            True

        AUTHOR:

        - Rob Beezer (2011-07-15)
        """
        # any nonzero entry in any matrix representation
        #   disqualifies the morphism as having totally zero outputs
        return self._matrix.is_zero()

    def is_equal_function(self, other) -> bool:
        r"""
        Determine if two morphisms are equal functions.

        INPUT:

        - ``other`` -- a morphism to compare with ``self``

        OUTPUT:

        Returns ``True`` precisely when the two morphisms have
        equal domains and codomains (as sets) and produce identical
        output when given the same input.  Otherwise returns ``False``.

        This is useful when ``self`` and ``other`` may have different
        representations.

        Sage's default comparison of matrix morphisms requires the
        domains to have the same bases and the codomains to have the
        same bases, and then compares the matrix representations.
        This notion of equality is more permissive (it will
        return ``True`` "more often"), but is more correct
        mathematically.

        EXAMPLES:

        Three morphisms defined by combinations of different
        bases for the domain and codomain and different functions.
        Two are equal, the third is different from both of the others.  ::

            sage: B = matrix(QQ, [[-3,  5, -4,  2],
            ....:                 [-1,  2, -1,  4],
            ....:                 [ 4, -6,  5, -1],
            ....:                 [-5,  7, -6,  1]])
            sage: U = (QQ^4).subspace_with_basis(B.rows())
            sage: C = matrix(QQ, [[-1, -6, -4],
            ....:                 [ 3, -5,  6],
            ....:                 [ 1,  2,  3]])
            sage: V = (QQ^3).subspace_with_basis(C.rows())
            sage: H = Hom(U, V)

            sage: D = matrix(QQ, [[-7, -2, -5,  2],
            ....:                 [-5,  1, -4, -8],
            ....:                 [ 1, -1,  1,  4],
            ....:                 [-4, -1, -3,   1]])
            sage: X = (QQ^4).subspace_with_basis(D.rows())
            sage: E = matrix(QQ, [[ 4, -1,  4],
            ....:                 [ 5, -4, -5],
            ....:                 [-1,  0, -2]])
            sage: Y = (QQ^3).subspace_with_basis(E.rows())
            sage: K = Hom(X, Y)

            sage: f = lambda x: vector(QQ, [x[0]+x[1], 2*x[1]-4*x[2], 5*x[3]])
            sage: g = lambda x: vector(QQ, [x[0]-x[2], 2*x[1]-4*x[2], 5*x[3]])

            sage: rho = H(f)
            sage: phi = K(f)
            sage: zeta = H(g)

            sage: rho.is_equal_function(phi)
            True
            sage: phi.is_equal_function(rho)
            True
            sage: zeta.is_equal_function(rho)
            False
            sage: phi.is_equal_function(zeta)
            False

        TESTS::

            sage: H = Hom(ZZ^2, ZZ^2)
            sage: phi = H(matrix(ZZ, 2, range(4)))
            sage: phi.is_equal_function('junk')
            Traceback (most recent call last):
            ...
            TypeError: can only compare to a matrix morphism, not junk

        AUTHOR:

        - Rob Beezer (2011-07-15)
        """
        if not isinstance(other, MatrixMorphism_abstract):
            msg = 'can only compare to a matrix morphism, not {0}'
            raise TypeError(msg.format(other))
        if self.domain() != other.domain():
            return False
        if self.codomain() != other.codomain():
            return False
        # check agreement on any basis of the domain
        return all(self(u) == other(u) for u in self.domain().basis())

    def restrict_domain(self, sub):
        """
        Restrict this matrix morphism to a subspace sub of the domain. The
        subspace sub should have a basis() method and elements of the basis
        should be coercible into domain.

        The resulting morphism has the same codomain as before, but a new
        domain.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([3*V.0, 2*V.1])
            sage: phi.restrict_domain(V.span([V.0]))
            Free module morphism defined by the matrix
            [3 0]
            Domain: Free module of degree 2 and rank 1 over Integer Ring
            Echelon ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...
            sage: phi.restrict_domain(V.span([V.1]))
            Free module morphism defined by the matrix
            [0 2]...
            sage: m = matrix(2, range(1,5))
            sage: f1 = V.hom(m); f2 = V.hom(m, side='right')
            sage: SV = V.span([V.0])
            sage: f1.restrict_domain(SV)
            Free module morphism defined by the matrix
            [1 2]...
            sage: f2.restrict_domain(SV)
            Free module morphism defined as left-multiplication by the matrix
            [1]
            [3]...
        """
        D = self.domain()
        if hasattr(D, 'coordinate_module'):
            # We only have to do this in case the module supports
            # alternative basis.  Some modules do, some modules don't.
            V = D.coordinate_module(sub)
        else:
            V = sub.free_module()
        if self.side() == "right":
            A = self.matrix().transpose().restrict_domain(V).transpose()
        else:
            A = self.matrix().restrict_domain(V)
        H = sub.Hom(self.codomain())
        try:
            return H(A, side=self.side())
        except Exception:
            return H(A)

    def restrict_codomain(self, sub):
        """
        Restrict this matrix morphism to a subspace sub of the codomain.

        The resulting morphism has the same domain as before, but a new
        codomain.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([4*(V.0+V.1),0])
            sage: W = V.span([2*(V.0+V.1)])
            sage: phi
            Free module morphism defined by the matrix
            [4 4]
            [0 0]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Ambient free module of rank 2 over the principal ideal domain ...
            sage: psi = phi.restrict_codomain(W); psi
            Free module morphism defined by the matrix
            [2]
            [0]
            Domain: Ambient free module of rank 2 over the principal ideal domain ...
            Codomain: Free module of degree 2 and rank 1 over Integer Ring
            Echelon ...
            sage: phi2 = phi.side_switch(); phi2.restrict_codomain(W)
            Free module morphism defined as left-multiplication by the matrix
            [2 0]
            Domain: Ambient free module of rank 2 over the principal ideal domain Integer Ring
            Codomain: Free module of degree 2 and rank 1 over Integer Ring
            Echelon ...

        An example in which the codomain equals the full ambient space, but
        with a different basis::

            sage: V = QQ^2
            sage: W = V.span_of_basis([[1,2],[3,4]])
            sage: phi = V.hom(matrix(QQ,2,[1,0,2,0]),W)
            sage: phi.matrix()
            [1 0]
            [2 0]
            sage: phi(V.0)
            (1, 2)
            sage: phi(V.1)
            (2, 4)
            sage: X = V.span([[1,2]]); X
            Vector space of degree 2 and dimension 1 over Rational Field
            Basis matrix:
            [1 2]
            sage: phi(V.0) in X
            True
            sage: phi(V.1) in X
            True
            sage: psi = phi.restrict_codomain(X); psi
            Vector space morphism represented by the matrix:
            [1]
            [2]
            Domain: Vector space of dimension 2 over Rational Field
            Codomain: Vector space of degree 2 and dimension 1 over Rational Field
            Basis matrix:
            [1 2]
            sage: psi(V.0)
            (1, 2)
            sage: psi(V.1)
            (2, 4)
            sage: psi(V.0).parent() is X
            True
        """
        H = self.domain().Hom(sub)
        C = self.codomain()
        if hasattr(C, 'coordinate_module'):
            # We only have to do this in case the module supports
            # alternative basis.  Some modules do, some modules don't.
            V = C.coordinate_module(sub)
        else:
            V = sub.free_module()
        try:
            if self.side() == "right":
                return H(self.matrix().transpose().restrict_codomain(V).transpose(), side='right')
            return H(self.matrix().restrict_codomain(V))
        except Exception:
            return H(self.matrix().restrict_codomain(V))

    def restrict(self, sub):
        """
        Restrict this matrix morphism to a subspace sub of the domain.

        The codomain and domain of the resulting matrix are both sub.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([3*V.0, 2*V.1])
            sage: phi.restrict(V.span([V.0]))
            Free module morphism defined by the matrix
            [3]
            Domain: Free module of degree 2 and rank 1 over Integer Ring
            Echelon ...
            Codomain: Free module of degree 2 and rank 1 over Integer Ring
            Echelon ...

            sage: V = (QQ^2).span_of_basis([[1,2], [3,4]])
            sage: phi = V.hom([V.0 + V.1, 2*V.1])
            sage: phi(V.1) == 2*V.1
            True
            sage: W = span([V.1])
            sage: phi(W)
            Vector space of degree 2 and dimension 1 over Rational Field
            Basis matrix:
            [  1 4/3]
            sage: psi = phi.restrict(W); psi
            Vector space morphism represented by the matrix:
            [2]
            Domain:   Vector space of degree 2 and dimension 1 over Rational Field
                      Basis matrix:
                      [  1 4/3]
            Codomain: Vector space of degree 2 and dimension 1 over Rational Field
                      Basis matrix:
                      [  1 4/3]
            sage: psi.domain() == W
            True
            sage: psi(W.0) == 2*W.0
            True

        ::

            sage: V = ZZ^3
            sage: h1 = V.hom([V.0, V.1+V.2, -V.1+V.2])
            sage: h2 = h1.side_switch()
            sage: SV = V.span([2*V.1,2*V.2])
            sage: h1.restrict(SV)
            Free module morphism defined by the matrix
            [ 1  1]
            [-1  1]
            Domain:   Free module of degree 3 and rank 2 over Integer Ring
                      Echelon basis matrix:
                      [0 2 0]
                      [0 0 2]
            Codomain: Free module of degree 3 and rank 2 over Integer Ring
                      Echelon basis matrix:
                      [0 2 0]
                      [0 0 2]
            sage: h2.restrict(SV)
            Free module morphism defined as left-multiplication by the matrix
            [ 1 -1]
            [ 1  1]
            Domain:   Free module of degree 3 and rank 2 over Integer Ring
                      Echelon basis matrix:
                      [0 2 0]
                      [0 0 2]
            Codomain: Free module of degree 3 and rank 2 over Integer Ring
                      Echelon basis matrix:
                      [0 2 0]
                      [0 0 2]
        """
        if not self.is_endomorphism():
            raise ArithmeticError("matrix morphism must be an endomorphism")
        D = self.domain()
        C = self.codomain()
        if D is not C and (D.basis() != C.basis()):
            # Tricky case when two bases for same space
            return self.restrict_domain(sub).restrict_codomain(sub)
        if hasattr(D, 'coordinate_module'):
            # We only have to do this in case the module supports
            # alternative basis.  Some modules do, some modules don't.
            V = D.coordinate_module(sub)
        else:
            V = sub.free_module()
        if self.side() == "right":
            A = self.matrix().transpose().restrict(V).transpose()
        else:
            A = self.matrix().restrict(V)
        H = sage.categories.homset.End(sub, self.domain().category())
        return H(A, side=self.side())


class MatrixMorphism(MatrixMorphism_abstract):
    """
    A morphism defined by a matrix.

    INPUT:

    - ``parent`` -- a homspace

    - ``A`` -- matrix or a :class:`MatrixMorphism_abstract` instance

    - ``copy_matrix`` -- boolean (default: ``True``); make an immutable copy of
      the matrix ``A`` if it is mutable. If ``False``, then this makes
      ``A`` immutable.
    """
    def __init__(self, parent, A, copy_matrix=True, side='left'):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.modules.matrix_morphism import MatrixMorphism
            sage: T = End(ZZ^3)
            sage: M = MatrixSpace(ZZ, 3)
            sage: I = M.identity_matrix()
            sage: A = MatrixMorphism(T, I)
            sage: loads(A.dumps()) == A
            True
        """
        if parent is None:
            raise ValueError("no parent given when creating this matrix morphism")
        if isinstance(A, MatrixMorphism_abstract):
            A = A.matrix()
        if side == "left":
            if A.nrows() != parent.domain().rank():
                raise ArithmeticError("number of rows of matrix (={}) must equal rank of domain (={})".format(A.nrows(), parent.domain().rank()))
            if A.ncols() != parent.codomain().rank():
                raise ArithmeticError("number of columns of matrix (={}) must equal rank of codomain (={})".format(A.ncols(), parent.codomain().rank()))
        if side == "right":
            if A.nrows() != parent.codomain().rank():
                raise ArithmeticError("number of rows of matrix (={}) must equal rank of codomain (={})".format(A.nrows(), parent.domain().rank()))
            if A.ncols() != parent.domain().rank():
                raise ArithmeticError("number of columns of matrix (={}) must equal rank of domain (={})".format(A.ncols(), parent.codomain().rank()))
        if A.is_mutable():
            if copy_matrix:
                from copy import copy
                A = copy(A)
            A.set_immutable()
        self._matrix = A
        MatrixMorphism_abstract.__init__(self, parent, side)

    def matrix(self, side=None):
        r"""
        Return a matrix that defines this morphism.

        INPUT:

        - ``side`` -- (default: ``None``) the side of the matrix
          where a vector is placed to effect the morphism (function)

        OUTPUT:

        A matrix which represents the morphism, relative to bases
        for the domain and codomain.  If the modules are provided
        with user bases, then the representation is relative to
        these bases.

        Internally, Sage represents a matrix morphism with the
        matrix multiplying a row vector placed to the left of the
        matrix.  If the option ``side='right'`` is used, then a
        matrix is returned that acts on a vector to the right of
        the matrix.  These two matrices are just transposes of
        each other and the difference is just a preference for
        the style of representation.

        EXAMPLES::

            sage: V = ZZ^2; W = ZZ^3
            sage: m = column_matrix([3*V.0 - 5*V.1, 4*V.0 + 2*V.1, V.0 + V.1])
            sage: phi = V.hom(m, W)
            sage: phi.matrix()
            [ 3  4  1]
            [-5  2  1]

            sage: phi.matrix(side='right')
            [ 3 -5]
            [ 4  2]
            [ 1  1]

        TESTS::

            sage: V = ZZ^2
            sage: phi = V.hom([3*V.0, 2*V.1])
            sage: phi.matrix(side='junk')
            Traceback (most recent call last):
            ...
            ValueError: side must be 'left' or 'right', not junk
        """
        if side not in ['left', 'right', None]:
            raise ValueError("side must be 'left' or 'right', not {}".format(side))
        if side == self.side() or side is None:
            return self._matrix
        return self._matrix.transpose()

    def is_injective(self) -> bool:
        """
        Tell whether ``self`` is injective.

        EXAMPLES::

            sage: V1 = QQ^2
            sage: V2 = QQ^3
            sage: phi = V1.hom(Matrix([[1,2,3], [4,5,6]]),V2)
            sage: phi.is_injective()
            True
            sage: psi = V2.hom(Matrix([[1,2], [3,4], [5,6]]),V1)
            sage: psi.is_injective()
            False

        AUTHOR:

        -- Simon King (2010-05)
        """
        if self.side() == 'left':
            ker = self._matrix.left_kernel()
        else:
            ker = self._matrix.right_kernel()
        return ker.dimension() == 0

    def is_surjective(self) -> bool:
        r"""
        Tell whether ``self`` is surjective.

        EXAMPLES::

            sage: V1 = QQ^2
            sage: V2 = QQ^3
            sage: phi = V1.hom(Matrix([[1,2,3], [4,5,6]]), V2)
            sage: phi.is_surjective()
            False
            sage: psi = V2.hom(Matrix([[1,2], [3,4], [5,6]]), V1)
            sage: psi.is_surjective()
            True

        An example over a PID that is not `\ZZ`.  ::

            sage: R.<x> = PolynomialRing(QQ)
            sage: A = R^2
            sage: B = R^2
            sage: H = A.hom([B([x^2 - 1, 1]), B([x^2, 1])])
            sage: H.image()
            Free module of degree 2 and rank 2 over Univariate Polynomial Ring in x over Rational Field
            Echelon basis matrix:
            [ 1  0]
            [ 0 -1]
            sage: H.is_surjective()
            True

        This tests if :issue:`11552` is fixed. ::

            sage: V = ZZ^2
            sage: m = matrix(ZZ, [[1,2], [0,2]])
            sage: phi = V.hom(m, V)
            sage: phi.lift(vector(ZZ, [0, 1]))
            Traceback (most recent call last):
            ...
            ValueError: element is not in the image
            sage: phi.is_surjective()
            False

        AUTHORS:

        - Simon King (2010-05)
        - Rob Beezer (2011-06-28)
        """
        # Testing equality of free modules over PIDs is unreliable
        #   see Issue #11579 for explanation and status
        # We test if image equals codomain with two inclusions
        #   reverse inclusion of below is trivially true
        return self.codomain().is_submodule(self.image())

    def _repr_(self):
        r"""
        Return string representation of this matrix morphism.

        This will typically be overloaded in a derived class.

        EXAMPLES::

            sage: V = ZZ^2; phi = V.hom([3*V.0, 2*V.1])
            sage: sage.modules.matrix_morphism.MatrixMorphism._repr_(phi)
            'Morphism defined by the matrix\n[3 0]\n[0 2]'

            sage: phi._repr_()
            'Free module morphism defined by the matrix\n[3 0]\n[0 2]\nDomain: Ambient free module of rank 2 over the principal ideal domain Integer Ring\nCodomain: Ambient free module of rank 2 over the principal ideal domain Integer Ring'
        """
        rep = "Morphism defined by the matrix\n{}".format(self.matrix())
        if self._side == 'right':
            rep += " acting by multiplication on the left"
        return rep
